import typing as t
from dataclasses import dataclass, field

from dataclasses_json.core import Json

from unstructured_ingest.connector.elasticsearch import (
    ElasticsearchDestinationConnector,
    ElasticsearchDocumentMeta,
    ElasticsearchIngestDoc,
    ElasticsearchIngestDocBatch,
    ElasticsearchSourceConnector,
    SimpleElasticsearchConfig,
)
from unstructured_ingest.enhanced_dataclass import enhanced_field
from unstructured_ingest.error import DestinationConnectionError, SourceConnectionError
from unstructured_ingest.interfaces import AccessConfig, BaseSingleIngestDoc
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.data_prep import flatten_dict, generator_batching_wbytes
from unstructured_ingest.utils.dep_check import requires_dependencies

if t.TYPE_CHECKING:
    from opensearchpy import OpenSearch

"""Since the actual OpenSearch project is a fork of Elasticsearch, we are relying
heavily on the Elasticsearch connector code, inheriting the functionality as much as possible."""


@dataclass
class OpenSearchAccessConfig(AccessConfig):
    hosts: t.Optional[t.List[str]] = None
    username: t.Optional[str] = None
    password: t.Optional[str] = enhanced_field(default=None, sensitive=True)
    use_ssl: bool = False
    verify_certs: bool = False
    ssl_show_warn: bool = False
    ca_certs: t.Optional[str] = None
    client_cert: t.Optional[str] = None
    client_key: t.Optional[str] = None

    def to_dict(self, **kwargs) -> t.Dict[str, Json]:
        d = super().to_dict(**kwargs)
        d["http_auth"] = (self.username, self.password)
        return d


@dataclass
class SimpleOpenSearchConfig(SimpleElasticsearchConfig):
    access_config: OpenSearchAccessConfig = None


@dataclass
class OpenSearchIngestDoc(ElasticsearchIngestDoc):
    """Class encapsulating fetching a doc and writing processed results (but not
    doing the processing!).

    Current implementation creates a python OpenSearch client to fetch each doc,
    rather than creating a client for each thread.
    """

    connector_config: SimpleOpenSearchConfig
    registry_name: str = "opensearch"

    @SourceConnectionError.wrap
    @requires_dependencies(["opensearchpy"], extras="opensearch")
    @BaseSingleIngestDoc.skip_if_file_exists
    def get_file(self):
        pass


@dataclass
class OpenSearchIngestDocBatch(ElasticsearchIngestDocBatch):
    connector_config: SimpleOpenSearchConfig
    ingest_docs: t.List[OpenSearchIngestDoc] = field(default_factory=list)
    registry_name: str = "opensearch_batch"

    @requires_dependencies(["opensearchpy"], extras="opensearch")
    def _get_docs(self):
        from opensearchpy import OpenSearch
        from opensearchpy.helpers import scan

        ops = OpenSearch(**self.connector_config.access_config.to_dict(apply_name_overload=False))
        scan_query = {
            "_source": self.connector_config.fields,
            "version": True,
            "query": {"ids": {"values": self.list_of_ids}},
        }

        result = scan(
            ops,
            query=scan_query,
            scroll="1m",
            index=self.connector_config.index_name,
        )
        return list(result)

    @SourceConnectionError.wrap
    @requires_dependencies(["opensearchpy"], extras="opensearch")
    def get_files(self):
        documents = self._get_docs()
        for doc in documents:
            ingest_doc = OpenSearchIngestDoc(
                processor_config=self.processor_config,
                read_config=self.read_config,
                connector_config=self.connector_config,
                document=doc,
                document_meta=ElasticsearchDocumentMeta(
                    self.connector_config.index_name, doc["_id"]
                ),
            )
            ingest_doc.update_source_metadata()
            doc_body = doc["_source"]
            filename = ingest_doc.filename
            flattened_dict = flatten_dict(dictionary=doc_body)
            str_values = [str(value) for value in flattened_dict.values()]
            concatenated_values = "\n".join(str_values)

            filename.parent.mkdir(parents=True, exist_ok=True)
            with open(filename, "w", encoding="utf8") as f:
                f.write(concatenated_values)
            self.ingest_docs.append(ingest_doc)


@dataclass
class OpenSearchSourceConnector(ElasticsearchSourceConnector):
    """Fetches particular fields from all documents in a given opensearch cluster and index"""

    connector_config: SimpleOpenSearchConfig
    _ops: t.Optional["OpenSearch"] = field(init=False, default=None)

    @property
    def ops(self):
        from opensearchpy import OpenSearch

        if self._ops is None:
            self._ops = OpenSearch(
                **self.connector_config.access_config.to_dict(apply_name_overload=False)
            )
        return self._ops

    def check_connection(self):
        try:
            assert self.ops.ping()
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise SourceConnectionError(f"failed to validate connection: {e}")

    @requires_dependencies(["opensearchpy"], extras="opensearch")
    def _get_doc_ids(self):
        """Fetches all document ids in an index"""
        from opensearchpy.helpers import scan

        hits = scan(
            self.ops,
            query=self.scan_query,
            scroll="1m",
            index=self.connector_config.index_name,
        )

        return [hit["_id"] for hit in hits]

    def get_ingest_docs(self):
        """Fetches all documents in an index, using ids that are fetched with _get_doc_ids"""
        ids = self._get_doc_ids()
        id_batches = [
            ids[
                i
                * self.connector_config.batch_size : (i + 1)  # noqa
                * self.connector_config.batch_size
            ]
            for i in range(
                (len(ids) + self.connector_config.batch_size - 1)
                // self.connector_config.batch_size
            )
        ]
        return [
            OpenSearchIngestDocBatch(
                connector_config=self.connector_config,
                processor_config=self.processor_config,
                read_config=self.read_config,
                list_of_ids=batched_ids,
            )
            for batched_ids in id_batches
        ]


@dataclass
class OpenSearchDestinationConnector(ElasticsearchDestinationConnector):
    connector_config: SimpleOpenSearchConfig
    _client: t.Optional["OpenSearch"] = field(init=False, default=None)

    @DestinationConnectionError.wrap
    @requires_dependencies(["opensearchpy"], extras="opensearch")
    def generate_client(self) -> "OpenSearch":
        from opensearchpy import OpenSearch

        return OpenSearch(**self.connector_config.access_config.to_dict(apply_name_overload=False))

    @requires_dependencies(["opensearchpy"], extras="opensearch")
    def write_dict(self, *args, elements_dict: t.List[t.Dict[str, t.Any]]) -> None:
        logger.info(
            f"writing document batches to destination"
            f" index named {self.connector_config.index_name}"
            f" at {self.connector_config.access_config.hosts}"
            f" with batch size (in bytes) {self.write_config.batch_size_bytes}"
            f" with {self.write_config.num_processes} (number of) processes"
        )
        from opensearchpy.helpers import parallel_bulk

        for batch in generator_batching_wbytes(
            elements_dict, batch_size_limit_bytes=self.write_config.batch_size_bytes
        ):
            for success, info in parallel_bulk(
                self.client, batch, thread_count=self.write_config.num_processes
            ):
                if not success:
                    logger.error(
                        "upload failed for a batch in opensearch destination connector:", info
                    )



from pathlib import Path

import pytest
from pytest_mock import MockerFixture

from unstructured_ingest.v2.processes.connectors.databricks.volumes_table import (
    DatabricksVolumeDeltaTableStager,
)


@pytest.fixture
def stager():
    return DatabricksVolumeDeltaTableStager()


@pytest.mark.parametrize(
    ("output_path", "called_output_path"),
    [
        (
            Path("/fake/path/output"),
            Path("/fake/path/output.json"),
        ),
        (
            Path("/fake/path/output.ndjson"),
            Path("/fake/path/output.json"),
        ),
    ],
)
def test_write_output(
    mocker: MockerFixture,
    stager: DatabricksVolumeDeltaTableStager,
    output_path: Path,
    called_output_path: Path,
):
    data = [{"key1": "value1", "key2": "value2"}]

    mock_get_data = mocker.patch(
        "unstructured_ingest.v2.processes.connectors.databricks.volumes_table.write_data",
        return_value=None,
    )

    stager.write_output(output_path, data)

    mock_get_data.assert_called_once_with(path=called_output_path, data=data, indent=None)



# https://developers.notion.com/reference/block#bulleted-list-item
from dataclasses import dataclass, field
from typing import List, Optional

from htmlBuilder.tags import HtmlTag, Li

from unstructured_ingest.connector.notion.interfaces import BlockBase
from unstructured_ingest.connector.notion.types.rich_text import RichText


@dataclass
class BulletedListItem(BlockBase):
    color: str
    children: List[dict] = field(default_factory=list)
    rich_text: List[RichText] = field(default_factory=list)

    @staticmethod
    def can_have_children() -> bool:
        return True

    @classmethod
    def from_dict(cls, data: dict):
        rich_text = data.pop("rich_text", [])
        return cls(
            color=data["color"],
            children=data.get("children", []),
            rich_text=[RichText.from_dict(rt) for rt in rich_text],
        )

    def get_html(self) -> Optional[HtmlTag]:
        return Li([], [rt.get_html() for rt in self.rich_text])



import click

from unstructured_ingest.v2.cli.base import DestCmd, SrcCmd
from unstructured_ingest.v2.processes.connector_registry import (
    destination_registry,
    source_registry,
)

src_cmds = [SrcCmd(cmd_name=k, registry_entry=v) for k, v in source_registry.items()]
dest_cmds = [DestCmd(cmd_name=k, registry_entry=v) for k, v in destination_registry.items()]

src: list[click.Group] = [v.get_cmd() for v in src_cmds]

dest: list[click.Command] = [v.get_cmd() for v in dest_cmds]



from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from pydantic import Field, SecretStr

from unstructured_ingest.embed.interfaces import (
    AsyncBaseEmbeddingEncoder,
    BaseEmbeddingEncoder,
    EmbeddingConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.errors import (
    RateLimitError as CustomRateLimitError,
)
from unstructured_ingest.v2.errors import UserAuthError, UserError, is_internal_error

if TYPE_CHECKING:
    from together import AsyncTogether, Together


class TogetherAIEmbeddingConfig(EmbeddingConfig):
    api_key: SecretStr
    embedder_model_name: str = Field(
        default="togethercomputer/m2-bert-80M-8k-retrieval", alias="model_name"
    )

    def wrap_error(self, e: Exception) -> Exception:
        if is_internal_error(e=e):
            return e
        # https://docs.together.ai/docs/error-codes
        from together.error import AuthenticationError, RateLimitError, TogetherException

        if not isinstance(e, TogetherException):
            logger.error(f"unhandled exception from openai: {e}", exc_info=True)
            return e
        message = e.args[0]
        if isinstance(e, AuthenticationError):
            return UserAuthError(message)
        if isinstance(e, RateLimitError):
            return CustomRateLimitError(message)
        return UserError(message)

    @requires_dependencies(["together"], extras="togetherai")
    def get_client(self) -> "Together":
        from together import Together

        return Together(api_key=self.api_key.get_secret_value())

    @requires_dependencies(["together"], extras="togetherai")
    def get_async_client(self) -> "AsyncTogether":
        from together import AsyncTogether

        return AsyncTogether(api_key=self.api_key.get_secret_value())


@dataclass
class TogetherAIEmbeddingEncoder(BaseEmbeddingEncoder):
    config: TogetherAIEmbeddingConfig

    def wrap_error(self, e: Exception) -> Exception:
        return self.config.wrap_error(e=e)

    def get_client(self) -> "Together":
        return self.config.get_client()

    def embed_batch(self, client: "Together", batch: list[str]) -> list[list[float]]:
        outputs = client.embeddings.create(model=self.config.embedder_model_name, input=batch)
        return [outputs.data[i].embedding for i in range(len(batch))]


@dataclass
class AsyncTogetherAIEmbeddingEncoder(AsyncBaseEmbeddingEncoder):
    config: TogetherAIEmbeddingConfig

    def wrap_error(self, e: Exception) -> Exception:
        return self.config.wrap_error(e=e)

    def get_client(self) -> "AsyncTogether":
        return self.config.get_async_client()

    async def embed_batch(self, client: Any, batch: list[str]) -> list[list[float]]:
        outputs = await client.embeddings.create(model=self.config.embedder_model_name, input=batch)
        return [outputs.data[i].embedding for i in range(len(batch))]



import os
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import pytest

from test.integration.connectors.utils.constants import SOURCE_TAG, UNCATEGORIZED_TAG
from test.integration.connectors.utils.validation.source import (
    SourceValidationConfigs,
    source_connector_validation,
)
from test.integration.utils import requires_env
from unstructured_ingest.error import SourceConnectionError
from unstructured_ingest.v2.processes.connectors.discord import (
    CONNECTOR_TYPE,
    DiscordAccessConfig,
    DiscordConnectionConfig,
    DiscordDownloader,
    DiscordDownloaderConfig,
    DiscordIndexer,
    DiscordIndexerConfig,
)


@dataclass(frozen=True)
class EnvData:
    token: Optional[str]
    channels: Optional[list[str]]


def get_env_data() -> EnvData:
    return EnvData(
        token=os.getenv("DISCORD_TOKEN"),
        channels=os.getenv("DISCORD_CHANNELS", default=[]).split(","),
    )


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, UNCATEGORIZED_TAG)
@requires_env("DISCORD_TOKEN", "DISCORD_CHANNELS")
async def test_discord_source():
    env = get_env_data()
    indexer_config = DiscordIndexerConfig(channels=env.channels)
    with tempfile.TemporaryDirectory() as tempdir:
        tempdir_path = Path(tempdir)
        connection_config = DiscordConnectionConfig(
            access_config=DiscordAccessConfig(token=env.token)
        )
        download_config = DiscordDownloaderConfig(download_dir=tempdir_path)
        indexer = DiscordIndexer(connection_config=connection_config, index_config=indexer_config)
        downloader = DiscordDownloader(
            connection_config=connection_config, download_config=download_config
        )
        expected_num_files = len(env.channels)
        await source_connector_validation(
            indexer=indexer,
            downloader=downloader,
            configs=SourceValidationConfigs(
                test_id=CONNECTOR_TYPE,
                expected_num_files=expected_num_files,
                expected_number_indexed_file_data=expected_num_files,
                validate_downloaded_files=True,
            ),
        )


@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, UNCATEGORIZED_TAG)
@requires_env("DISCORD_CHANNELS")
def test_discord_source_precheck_fail_no_token():
    indexer_config = DiscordIndexerConfig(channels=get_env_data().channels)

    connection_config = DiscordConnectionConfig(access_config=DiscordAccessConfig(token=""))
    indexer = DiscordIndexer(connection_config=connection_config, index_config=indexer_config)
    with pytest.raises(SourceConnectionError):
        indexer.precheck()


@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, UNCATEGORIZED_TAG)
@requires_env("DISCORD_TOKEN")
def test_discord_source_precheck_fail_no_channels():
    indexer_config = DiscordIndexerConfig(channels=[])

    connection_config = DiscordConnectionConfig(
        access_config=DiscordAccessConfig(token=get_env_data().token)
    )
    indexer = DiscordIndexer(connection_config=connection_config, index_config=indexer_config)
    with pytest.raises(SourceConnectionError):
        indexer.precheck()



from pathlib import Path

import pytest

int_test_dir = Path(__file__).parent
assets_dir = int_test_dir / "assets"


@pytest.fixture
def duckdb_schema() -> Path:
    schema_file = assets_dir / "duckdb-schema.sql"
    assert schema_file.exists()
    assert schema_file.is_file()
    return schema_file



from __future__ import annotations

import json
from dataclasses import InitVar, fields
from typing import Any, Callable, Optional, Type, TypeVar, Union

import dataclasses_json.core as dataclasses_json_core
from dataclasses_json import DataClassJsonMixin

from unstructured_ingest.enhanced_dataclass.core import _asdict

A = TypeVar("A", bound="EnhancedDataClassJsonMixin")

# Monkey-patch _decode_dataclass class to support name override
og_decode_dataclass = dataclasses_json_core._decode_dataclass


def custom_decode_dataclass(cls, kvs, infer_missing):
    dataclass_fields = fields(cls)
    for f in [
        field
        for field in dataclass_fields
        if hasattr(field, "overload_name") and getattr(field, "overload_name", None)
    ]:
        field_name = f.name
        overload_name = getattr(f, "overload_name")
        if isinstance(kvs, dict) and overload_name in kvs:
            kvs[field_name] = kvs.pop(overload_name)
    return og_decode_dataclass(cls, kvs, infer_missing)


dataclasses_json_core._decode_dataclass = custom_decode_dataclass


class EnhancedDataClassJsonMixin(DataClassJsonMixin):
    """A mixin class extending DataClassJsonMixin.

    This class extends the functionality of DataClassJsonMixin to provide enhanced functionality
    for JSON serialization and deserialization. It introduces options for redacting sensitive
    information, custom encoding, and more advanced schema handling.

    Attributes:
        N/A (No additional attributes)

    Methods:
        to_json: Serialize the object to JSON format with customizable options.
        from_dict: Deserialize a dictionary into an object of this class.
        to_dict: Convert the object to a dictionary with customizable options.
        schema: Generate a schema for validating and parsing JSON data based on this class.
    """

    @classmethod
    def check_init_var(cls):
        ann = cls.__dict__.get("__annotations__", {})
        init_vars = {k: v for k, v in ann.items() if isinstance(v, InitVar)}
        if init_vars:
            raise TypeError(
                "Class {} has the following fields defined with an InitVar which "
                "cannot be used with EnhancedDataClassJsonMixin: {}".format(
                    cls.__name__, ", ".join(init_vars.keys())
                )
            )

    def to_json(
        self,
        *,
        skipkeys: bool = False,
        ensure_ascii: bool = True,
        check_circular: bool = True,
        allow_nan: bool = True,
        indent: Optional[Union[int, str]] = None,
        separators: Optional[tuple[str, str]] = None,
        default: Optional[Callable[..., Any]] = None,
        sort_keys: bool = False,
        redact_sensitive: bool = False,
        redacted_text: str = "***REDACTED***",
        apply_name_overload: bool = True,
        **kw: Any,
    ) -> str:
        self.check_init_var()
        return json.dumps(
            self.to_dict(
                encode_json=False,
                redact_sensitive=redact_sensitive,
                redacted_text=redacted_text,
                apply_name_overload=apply_name_overload,
            ),
            cls=dataclasses_json_core._ExtendedEncoder,
            skipkeys=skipkeys,
            ensure_ascii=ensure_ascii,
            check_circular=check_circular,
            allow_nan=allow_nan,
            indent=indent,
            separators=separators,
            default=default,
            sort_keys=sort_keys,
            **kw,
        )

    @classmethod
    def from_dict(
        cls: Type[A],
        kvs: dataclasses_json_core.Json,
        *,
        infer_missing=False,
        apply_name_overload=False,
    ) -> A:
        cls.check_init_var()
        return dataclasses_json_core._decode_dataclass(cls, kvs, infer_missing)

    def to_dict(
        self,
        encode_json: bool = False,
        redact_sensitive: bool = False,
        redacted_text: str = "***REDACTED***",
        apply_name_overload: bool = True,
    ) -> dict[str, dataclasses_json_core.Json]:
        self.check_init_var()
        return _asdict(
            self,
            encode_json=encode_json,
            redact_sensitive=redact_sensitive,
            redacted_text=redacted_text,
            apply_name_overload=apply_name_overload,
        )



# https://developers.notion.com/reference/block#code
from dataclasses import dataclass, field
from typing import List, Optional

from htmlBuilder.tags import Br, Div, HtmlTag
from htmlBuilder.tags import Code as HtmlCode

from unstructured_ingest.connector.notion.interfaces import BlockBase
from unstructured_ingest.connector.notion.types.rich_text import RichText


@dataclass
class Code(BlockBase):
    language: str
    rich_text: List[RichText] = field(default_factory=list)
    caption: List[RichText] = field(default_factory=list)

    @staticmethod
    def can_have_children() -> bool:
        return False

    @classmethod
    def from_dict(cls, data: dict):
        rich_text = data.pop("rich_text", [])
        caption = data.pop("caption", [])
        return cls(
            language=data["language"],
            rich_text=[RichText.from_dict(rt) for rt in rich_text],
            caption=[RichText.from_dict(c) for c in caption],
        )

    def get_html(self) -> Optional[HtmlTag]:
        texts = []
        if self.rich_text:
            texts.append(HtmlCode([], [rt.get_html() for rt in self.rich_text]))
        if self.caption:
            texts.append(Div([], [rt.get_html() for rt in self.caption]))
        if not texts:
            return None
        joined = [Br()] * (len(texts) * 2 - 1)
        joined[0::2] = texts

        return Div([], joined)



import datetime as dt
import typing as t
from dataclasses import dataclass
from pathlib import Path

from unstructured_ingest.enhanced_dataclass import enhanced_field
from unstructured_ingest.error import SourceConnectionError, SourceConnectionNetworkError
from unstructured_ingest.interfaces import (
    AccessConfig,
    BaseConnectorConfig,
    BaseSingleIngestDoc,
    BaseSourceConnector,
    IngestDocCleanupMixin,
    SourceConnectorCleanupMixin,
    SourceMetadata,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.dep_check import (
    requires_dependencies,
)


@dataclass
class DiscordAccessConfig(AccessConfig):
    token: str = enhanced_field(sensitive=True)


@dataclass
class SimpleDiscordConfig(BaseConnectorConfig):
    """Connector config where channels is a comma separated list of
    Discord channels to pull messages from.
    """

    # Discord Specific Options
    access_config: DiscordAccessConfig
    channels: t.List[str]
    period: t.Optional[int] = None


@dataclass
class DiscordIngestDoc(IngestDocCleanupMixin, BaseSingleIngestDoc):
    """Class encapsulating fetching a doc and writing processed results (but not
    doing the processing!).
    Also includes a cleanup method. When things go wrong and the cleanup
    method is not called, the file is left behind on the filesystem to assist debugging.
    """

    connector_config: SimpleDiscordConfig
    channel: str
    days: t.Optional[int] = None
    registry_name: str = "discord"

    # NOTE(crag): probably doesn't matter,  but intentionally not defining tmp_download_file
    # __post_init__ for multiprocessing simplicity (no Path objects in initially
    # instantiated object)
    def _tmp_download_file(self):
        channel_file = self.channel + ".txt"
        return Path(self.read_config.download_dir) / channel_file

    @property
    def _output_filename(self):
        output_file = self.channel + ".json"
        return Path(self.processor_config.output_dir) / output_file

    def _create_full_tmp_dir_path(self):
        self._tmp_download_file().parent.mkdir(parents=True, exist_ok=True)

    @SourceConnectionNetworkError.wrap
    @requires_dependencies(dependencies=["discord"], extras="discord")
    def _get_messages(self):
        """Actually fetches the data from discord."""
        import discord
        from discord.ext import commands

        messages: t.List[discord.Message] = []
        jumpurl: t.List[str] = []
        intents = discord.Intents.default()
        intents.message_content = True
        bot = commands.Bot(command_prefix=">", intents=intents)

        @bot.event
        async def on_ready():
            try:
                after_date = None
                if self.days:
                    after_date = dt.datetime.utcnow() - dt.timedelta(days=self.days)
                channel = bot.get_channel(int(self.channel))
                jumpurl.append(channel.jump_url)  # type: ignore
                async for msg in channel.history(after=after_date):  # type: ignore
                    messages.append(msg)
                await bot.close()
            except Exception:
                logger.error("Error fetching messages")
                await bot.close()
                raise

        bot.run(self.connector_config.access_config.token)
        jump_url = None if len(jumpurl) < 1 else jumpurl[0]
        return messages, jump_url

    def update_source_metadata(self, **kwargs):
        messages, jump_url = kwargs.get("messages_tuple", self._get_messages())
        if messages == []:
            self.source_metadata = SourceMetadata(
                exists=False,
            )
            return
        dates = [m.created_at for m in messages if m.created_at]
        dates.sort()
        self.source_metadata = SourceMetadata(
            date_created=dates[0].isoformat(),
            date_modified=dates[-1].isoformat(),
            source_url=jump_url,
            exists=True,
        )

    @SourceConnectionError.wrap
    @BaseSingleIngestDoc.skip_if_file_exists
    def get_file(self):
        self._create_full_tmp_dir_path()

        messages, jump_url = self._get_messages()
        self.update_source_metadata(messages_tuple=(messages, jump_url))
        if messages == []:
            raise ValueError(f"Failed to retrieve messages from Discord channel {self.channel}")
        self._tmp_download_file().parent.mkdir(parents=True, exist_ok=True)
        with open(self._tmp_download_file(), "w") as f:
            for m in messages:
                f.write(m.content + "\n")

    @property
    def filename(self):
        """The filename of the file created from a discord channel"""
        return self._tmp_download_file()

    @property
    def version(self) -> t.Optional[str]:
        return None

    @property
    def record_locator(self) -> t.Optional[t.Dict[str, t.Any]]:
        return {
            "channel": self.channel,
        }


class DiscordSourceConnector(SourceConnectorCleanupMixin, BaseSourceConnector):
    """Objects of this class support fetching document(s) from"""

    connector_config: SimpleDiscordConfig

    def initialize(self):
        pass

    @requires_dependencies(dependencies=["discord"], extras="discord")
    def check_connection(self):
        import asyncio

        import discord
        from discord.client import Client

        intents = discord.Intents.default()
        try:
            client = Client(intents=intents)
            asyncio.run(client.start(token=self.connector_config.access_config.token))
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise SourceConnectionError(f"failed to validate connection: {e}")

    def get_ingest_docs(self):
        return [
            DiscordIngestDoc(
                connector_config=self.connector_config,
                processor_config=self.processor_config,
                read_config=self.read_config,
                channel=channel,
                days=self.connector_config.period,
            )
            for channel in self.connector_config.channels
        ]



from __future__ import annotations

import unstructured_ingest.v2.processes.connectors.databricks  # noqa: F401
import unstructured_ingest.v2.processes.connectors.duckdb  # noqa: F401
import unstructured_ingest.v2.processes.connectors.elasticsearch  # noqa: F401
import unstructured_ingest.v2.processes.connectors.fsspec  # noqa: F401
import unstructured_ingest.v2.processes.connectors.kafka  # noqa: F401
import unstructured_ingest.v2.processes.connectors.lancedb  # noqa: F401
import unstructured_ingest.v2.processes.connectors.qdrant  # noqa: F401
import unstructured_ingest.v2.processes.connectors.sql  # noqa: F401
import unstructured_ingest.v2.processes.connectors.weaviate  # noqa: F401
from unstructured_ingest.v2.processes.connector_registry import (
    add_destination_entry,
    add_source_entry,
)

from .airtable import CONNECTOR_TYPE as AIRTABLE_CONNECTOR_TYPE
from .airtable import airtable_source_entry
from .astradb import CONNECTOR_TYPE as ASTRA_DB_CONNECTOR_TYPE
from .astradb import astra_db_destination_entry, astra_db_source_entry
from .azure_ai_search import CONNECTOR_TYPE as AZURE_AI_SEARCH_CONNECTOR_TYPE
from .azure_ai_search import azure_ai_search_destination_entry
from .chroma import CONNECTOR_TYPE as CHROMA_CONNECTOR_TYPE
from .chroma import chroma_destination_entry
from .confluence import CONNECTOR_TYPE as CONFLUENCE_CONNECTOR_TYPE
from .confluence import confluence_source_entry
from .couchbase import CONNECTOR_TYPE as COUCHBASE_CONNECTOR_TYPE
from .couchbase import couchbase_destination_entry, couchbase_source_entry
from .delta_table import CONNECTOR_TYPE as DELTA_TABLE_CONNECTOR_TYPE
from .delta_table import delta_table_destination_entry
from .discord import CONNECTOR_TYPE as DISCORD_CONNECTOR_TYPE
from .discord import discord_source_entry
from .gitlab import CONNECTOR_TYPE as GITLAB_CONNECTOR_TYPE
from .gitlab import gitlab_source_entry
from .google_drive import CONNECTOR_TYPE as GOOGLE_DRIVE_CONNECTOR_TYPE
from .google_drive import google_drive_source_entry
from .kdbai import CONNECTOR_TYPE as KDBAI_CONNECTOR_TYPE
from .kdbai import kdbai_destination_entry
from .local import CONNECTOR_TYPE as LOCAL_CONNECTOR_TYPE
from .local import local_destination_entry, local_source_entry
from .milvus import CONNECTOR_TYPE as MILVUS_CONNECTOR_TYPE
from .milvus import milvus_destination_entry
from .mongodb import CONNECTOR_TYPE as MONGODB_CONNECTOR_TYPE
from .mongodb import mongodb_destination_entry, mongodb_source_entry
from .neo4j import CONNECTOR_TYPE as NEO4J_CONNECTOR_TYPE
from .neo4j import neo4j_destination_entry
from .notion.connector import CONNECTOR_TYPE as NOTION_CONNECTOR_TYPE
from .notion.connector import notion_source_entry
from .onedrive import CONNECTOR_TYPE as ONEDRIVE_CONNECTOR_TYPE
from .onedrive import onedrive_destination_entry, onedrive_source_entry
from .outlook import CONNECTOR_TYPE as OUTLOOK_CONNECTOR_TYPE
from .outlook import outlook_source_entry
from .pinecone import CONNECTOR_TYPE as PINECONE_CONNECTOR_TYPE
from .pinecone import pinecone_destination_entry
from .redisdb import CONNECTOR_TYPE as REDIS_CONNECTOR_TYPE
from .redisdb import redis_destination_entry
from .salesforce import CONNECTOR_TYPE as SALESFORCE_CONNECTOR_TYPE
from .salesforce import salesforce_source_entry
from .sharepoint import CONNECTOR_TYPE as SHAREPOINT_CONNECTOR_TYPE
from .sharepoint import sharepoint_source_entry
from .slack import CONNECTOR_TYPE as SLACK_CONNECTOR_TYPE
from .slack import slack_source_entry
from .vectara import CONNECTOR_TYPE as VECTARA_CONNECTOR_TYPE
from .vectara import vectara_destination_entry

add_source_entry(source_type=ASTRA_DB_CONNECTOR_TYPE, entry=astra_db_source_entry)
add_destination_entry(destination_type=ASTRA_DB_CONNECTOR_TYPE, entry=astra_db_destination_entry)

add_destination_entry(destination_type=CHROMA_CONNECTOR_TYPE, entry=chroma_destination_entry)

add_source_entry(source_type=COUCHBASE_CONNECTOR_TYPE, entry=couchbase_source_entry)
add_destination_entry(destination_type=COUCHBASE_CONNECTOR_TYPE, entry=couchbase_destination_entry)

add_destination_entry(
    destination_type=DELTA_TABLE_CONNECTOR_TYPE, entry=delta_table_destination_entry
)


add_source_entry(source_type=GOOGLE_DRIVE_CONNECTOR_TYPE, entry=google_drive_source_entry)

add_source_entry(source_type=LOCAL_CONNECTOR_TYPE, entry=local_source_entry)
add_destination_entry(destination_type=LOCAL_CONNECTOR_TYPE, entry=local_destination_entry)

add_source_entry(source_type=ONEDRIVE_CONNECTOR_TYPE, entry=onedrive_source_entry)
add_destination_entry(destination_type=ONEDRIVE_CONNECTOR_TYPE, entry=onedrive_destination_entry)

add_destination_entry(destination_type=NEO4J_CONNECTOR_TYPE, entry=neo4j_destination_entry)

add_source_entry(source_type=SALESFORCE_CONNECTOR_TYPE, entry=salesforce_source_entry)

add_destination_entry(destination_type=MONGODB_CONNECTOR_TYPE, entry=mongodb_destination_entry)
add_source_entry(source_type=MONGODB_CONNECTOR_TYPE, entry=mongodb_source_entry)

add_destination_entry(destination_type=PINECONE_CONNECTOR_TYPE, entry=pinecone_destination_entry)
add_source_entry(source_type=SHAREPOINT_CONNECTOR_TYPE, entry=sharepoint_source_entry)

add_destination_entry(destination_type=MILVUS_CONNECTOR_TYPE, entry=milvus_destination_entry)
add_destination_entry(
    destination_type=AZURE_AI_SEARCH_CONNECTOR_TYPE,
    entry=azure_ai_search_destination_entry,
)

add_destination_entry(destination_type=KDBAI_CONNECTOR_TYPE, entry=kdbai_destination_entry)
add_source_entry(source_type=AIRTABLE_CONNECTOR_TYPE, entry=airtable_source_entry)
add_source_entry(source_type=NOTION_CONNECTOR_TYPE, entry=notion_source_entry)

add_source_entry(source_type=OUTLOOK_CONNECTOR_TYPE, entry=outlook_source_entry)

add_source_entry(source_type=GITLAB_CONNECTOR_TYPE, entry=gitlab_source_entry)

add_source_entry(source_type=SLACK_CONNECTOR_TYPE, entry=slack_source_entry)

add_destination_entry(destination_type=VECTARA_CONNECTOR_TYPE, entry=vectara_destination_entry)
add_source_entry(source_type=CONFLUENCE_CONNECTOR_TYPE, entry=confluence_source_entry)

add_source_entry(source_type=DISCORD_CONNECTOR_TYPE, entry=discord_source_entry)
add_destination_entry(destination_type=REDIS_CONNECTOR_TYPE, entry=redis_destination_entry)



"""Defines Abstract Base Classes (ABC's) core to batch processing documents
through Unstructured."""

from __future__ import annotations

import functools
import json
import os
import re
from abc import ABC, abstractmethod
from dataclasses import InitVar, dataclass, field
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Type, TypeVar

from dataclasses_json import DataClassJsonMixin
from dataclasses_json.core import Json, _decode_dataclass

from unstructured_ingest.enhanced_dataclass import EnhancedDataClassJsonMixin, enhanced_field
from unstructured_ingest.enhanced_dataclass.core import _asdict
from unstructured_ingest.error import PartitionError, SourceConnectionError
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.data_prep import flatten_dict
from unstructured_ingest.v2.unstructured_api import call_api

if TYPE_CHECKING:
    from unstructured.documents.elements import Element

    from unstructured_ingest.embed.interfaces import BaseEmbeddingEncoder

A = TypeVar("A", bound="DataClassJsonMixin")

# -- Needed to resolve TypeError raised by using InitVar and __future__.annotations
# -- See more here: https://stackoverflow.com/questions/70400639/
InitVar.__call__ = lambda *args: None  # type: ignore

SUPPORTED_REMOTE_FSSPEC_PROTOCOLS = [
    "s3",
    "s3a",
    "abfs",
    "az",
    "gs",
    "gcs",
    "box",
    "dropbox",
    "sftp",
]


@dataclass
class BaseSessionHandle(ABC):
    """Abstract Base Class for sharing resources that are local to an individual process.
    e.g., a connection for making a request for fetching documents."""


@dataclass
class BaseConfig(EnhancedDataClassJsonMixin, ABC):
    pass


@dataclass
class AccessConfig(BaseConfig):
    """Meant to designate holding any sensitive information associated with other configs
    and also for access specific configs."""


@dataclass
class RetryStrategyConfig(BaseConfig):
    """
    Contains all info needed for decorator to pull from `self` for backoff
    and retry triggered by exception.

    Args:
        max_retries: The maximum number of attempts to make before giving
            up. Once exhausted, the exception will be allowed to escape.
            The default value of None means there is no limit to the
            number of tries. If a callable is passed, it will be
            evaluated at runtime and its return value used.
        max_retry_time: The maximum total amount of time to try for before
            giving up. Once expired, the exception will be allowed to
            escape. If a callable is passed, it will be
            evaluated at runtime and its return value used.
    """

    max_retries: Optional[int] = None
    max_retry_time: Optional[float] = None


@dataclass
class PartitionConfig(BaseConfig):
    # where to write structured data outputs
    pdf_infer_table_structure: bool = False
    strategy: str = "auto"
    ocr_languages: Optional[list[str]] = None
    encoding: Optional[str] = None
    additional_partition_args: dict[str, Any] = field(default_factory=dict)
    skip_infer_table_types: Optional[list[str]] = None
    fields_include: list[str] = field(
        default_factory=lambda: ["element_id", "text", "type", "metadata", "embeddings"],
    )
    flatten_metadata: bool = False
    metadata_exclude: list[str] = field(default_factory=list)
    metadata_include: list[str] = field(default_factory=list)
    partition_endpoint: Optional[str] = "https://api.unstructuredapp.io/general/v0/general"
    partition_by_api: bool = False
    api_key: Optional[str] = str(enhanced_field(default=None, sensitive=True)) or None
    hi_res_model_name: Optional[str] = None


@dataclass
class ProcessorConfig(BaseConfig):
    reprocess: bool = False
    verbose: bool = False
    work_dir: str = str((Path.home() / ".cache" / "unstructured" / "ingest" / "pipeline").resolve())
    output_dir: str = "structured-output"
    num_processes: int = 2
    raise_on_error: bool = False


@dataclass
class FileStorageConfig(BaseConfig):
    remote_url: str
    uncompress: bool = False
    recursive: bool = False
    file_glob: Optional[list[str]] = None


@dataclass
class FsspecConfig(FileStorageConfig):
    access_config: Optional[AccessConfig] = None
    protocol: str = field(init=False)
    path_without_protocol: str = field(init=False)
    dir_path: str = field(init=False)
    file_path: str = field(init=False)

    def get_access_config(self) -> dict[str, Any]:
        if self.access_config:
            return self.access_config.to_dict(apply_name_overload=False)
        else:
            return {}

    def __post_init__(self):
        self.protocol, self.path_without_protocol = self.remote_url.split("://")
        if self.protocol not in SUPPORTED_REMOTE_FSSPEC_PROTOCOLS:
            raise ValueError(
                f"Protocol {self.protocol} not supported yet, only "
                f"{SUPPORTED_REMOTE_FSSPEC_PROTOCOLS} are supported.",
            )

        # dropbox root is an empty string
        match = re.match(rf"{self.protocol}://([\s])/", self.remote_url)
        if match and self.protocol == "dropbox":
            self.dir_path = " "
            self.file_path = ""
            return

        # dropbox paths can start with slash
        match = re.match(rf"{self.protocol}:///([^/\s]+?)/([^\s]*)", self.remote_url)
        if match and self.protocol == "dropbox":
            self.dir_path = match.group(1)
            self.file_path = match.group(2) or ""
            return

        # just a path with no trailing prefix
        match = re.match(rf"{self.protocol}://([^/\s]+?)(/*)$", self.remote_url)
        if match:
            self.dir_path = match.group(1)
            self.file_path = ""
            return

        # valid path with a dir and/or file
        match = re.match(rf"{self.protocol}://([^/\s]+?)/([^\s]*)", self.remote_url)
        if not match:
            raise ValueError(
                f"Invalid path {self.remote_url}. "
                f"Expected <protocol>://<dir-path>/<file-or-dir-path>.",
            )
        self.dir_path = match.group(1)
        self.file_path = match.group(2) or ""


@dataclass
class ReadConfig(BaseConfig):
    # where raw documents are stored for processing, and then removed if not preserve_downloads
    download_dir: Optional[str] = ""
    re_download: bool = False
    preserve_downloads: bool = False
    download_only: bool = False
    max_docs: Optional[int] = None


@dataclass
class EmbeddingConfig(BaseConfig):
    provider: str
    api_key: Optional[str] = str(enhanced_field(default=None, sensitive=True)) or None
    model_name: Optional[str] = None
    aws_access_key_id: Optional[str] = None
    aws_secret_access_key: Optional[str] = None
    aws_region: Optional[str] = None

    def get_embedder(self) -> "BaseEmbeddingEncoder":
        kwargs: dict[str, Any] = {}
        if self.api_key:
            kwargs["api_key"] = self.api_key
        if self.model_name:
            kwargs["model_name"] = self.model_name
        # TODO make this more dynamic to map to encoder configs
        if self.provider == "openai":
            from unstructured_ingest.embed.openai import (
                OpenAIEmbeddingConfig,
                OpenAIEmbeddingEncoder,
            )

            return OpenAIEmbeddingEncoder(config=OpenAIEmbeddingConfig(**kwargs))
        elif self.provider == "huggingface":
            from unstructured_ingest.embed.huggingface import (
                HuggingFaceEmbeddingConfig,
                HuggingFaceEmbeddingEncoder,
            )

            return HuggingFaceEmbeddingEncoder(config=HuggingFaceEmbeddingConfig(**kwargs))
        elif self.provider == "octoai":
            from unstructured_ingest.embed.octoai import (
                OctoAiEmbeddingConfig,
                OctoAIEmbeddingEncoder,
            )

            return OctoAIEmbeddingEncoder(config=OctoAiEmbeddingConfig(**kwargs))
        elif self.provider == "bedrock":
            from unstructured_ingest.embed.bedrock import (
                BedrockEmbeddingConfig,
                BedrockEmbeddingEncoder,
            )

            return BedrockEmbeddingEncoder(
                config=BedrockEmbeddingConfig(
                    aws_access_key_id=self.aws_access_key_id,
                    aws_secret_access_key=self.aws_secret_access_key,
                    region_name=self.aws_region,
                )
            )
        elif self.provider == "vertexai":
            from unstructured_ingest.embed.vertexai import (
                VertexAIEmbeddingConfig,
                VertexAIEmbeddingEncoder,
            )

            return VertexAIEmbeddingEncoder(config=VertexAIEmbeddingConfig(**kwargs))
        elif self.provider == "voyageai":
            from unstructured_ingest.embed.voyageai import (
                VoyageAIEmbeddingConfig,
                VoyageAIEmbeddingEncoder,
            )

            return VoyageAIEmbeddingEncoder(config=VoyageAIEmbeddingConfig(**kwargs))
        else:
            raise ValueError(f"{self.provider} not a recognized encoder")


@dataclass
class ChunkingConfig(BaseConfig):
    chunk_elements: InitVar[bool] = False
    chunking_strategy: Optional[str] = None
    combine_text_under_n_chars: Optional[int] = None
    include_orig_elements: Optional[bool] = None
    max_characters: Optional[int] = None
    multipage_sections: Optional[bool] = None
    new_after_n_chars: Optional[int] = None
    overlap: Optional[int] = None
    overlap_all: Optional[bool] = None

    def __post_init__(self, chunk_elements: bool) -> None:
        """Resolve chunking_strategy if chunk_elements is True.

        If chunk_elements is True and chunking_strategy is None, default to 'by_title'. Otherwise,
        do nothing and keep the defined value of chunking_strategy."
        """
        if chunk_elements and self.chunking_strategy is None:
            self.chunking_strategy = "by_title"


@dataclass
class PermissionsConfig(BaseConfig):
    application_id: Optional[str] = enhanced_field(overload_name="permissions_application_id")
    tenant: Optional[str] = enhanced_field(overload_name="permissions_tenant")
    client_cred: Optional[str] = enhanced_field(
        default=None, sensitive=True, overload_name="permissions_client_cred"
    )


# module-level variable to store session handle
global_write_session_handle: Optional[BaseSessionHandle] = None


@dataclass
class WriteConfig(BaseConfig):
    pass


@dataclass
class BaseConnectorConfig(BaseConfig, ABC):
    """Abstract definition on which to define connector-specific attributes."""


@dataclass
class SourceMetadata(EnhancedDataClassJsonMixin, ABC):
    date_created: Optional[str] = None
    date_modified: Optional[str] = None
    version: Optional[str] = None
    source_url: Optional[str] = None
    exists: Optional[bool] = None
    permissions_data: Optional[list[dict[str, Any]]] = None


class IngestDocJsonMixin(EnhancedDataClassJsonMixin):
    """
    Inherently, DataClassJsonMixin does not add in any @property fields to the json/dict
    created from the dataclass. This explicitly sets properties to look for on the IngestDoc
    class when creating the json/dict for serialization purposes.
    """

    metadata_properties = [
        "date_created",
        "date_modified",
        "date_processed",
        "exists",
        "permissions_data",
        "version",
        "source_url",
    ]
    properties_to_serialize = [
        "base_filename",
        "filename",
        "_output_filename",
        "record_locator",
        "_source_metadata",
        "unique_id",
    ]

    def add_props(self, as_dict: dict[str, Any], props: list[str]):
        for prop in props:
            val = getattr(self, prop)
            if isinstance(val, Path):
                val = str(val)
            if isinstance(val, DataClassJsonMixin):
                val = val.to_dict(encode_json=False)
            as_dict[prop] = val

    def to_dict(self, **kwargs) -> dict[str, Json]:
        as_dict = _asdict(self, **kwargs)
        if "_session_handle" in as_dict:
            as_dict.pop("_session_handle", None)
        self.add_props(as_dict=as_dict, props=self.properties_to_serialize)
        if getattr(self, "_source_metadata") is not None:
            self.add_props(as_dict=as_dict, props=self.metadata_properties)
        return as_dict

    @classmethod
    def from_dict(
        cls: Type[A], kvs: Json, *, infer_missing=False, apply_name_overload: bool = True
    ) -> A:
        doc = super().from_dict(
            kvs=kvs, infer_missing=infer_missing, apply_name_overload=apply_name_overload
        )
        if meta := kvs.get("_source_metadata"):
            setattr(doc, "_source_metadata", SourceMetadata.from_dict(meta))
        if date_processed := kvs.get("_date_processed"):
            setattr(doc, "_date_processed", date_processed)
        return doc


class BatchIngestDocJsonMixin(EnhancedDataClassJsonMixin):
    """
    Inherently, DataClassJsonMixin does not add in any @property fields to the json/dict
    created from the dataclass. This explicitly sets properties to look for on the IngestDoc
    class when creating the json/dict for serialization purposes.
    """

    properties_to_serialize = ["unique_id"]

    def add_props(self, as_dict: dict[str, Any], props: list[str]):
        for prop in props:
            val = getattr(self, prop)
            if isinstance(val, Path):
                val = str(val)
            if isinstance(val, DataClassJsonMixin):
                val = val.to_dict(encode_json=False)
            as_dict[prop] = val

    def to_dict(self, encode_json=False) -> dict[str, Json]:
        as_dict = _asdict(self, encode_json=encode_json)
        self.add_props(as_dict=as_dict, props=self.properties_to_serialize)
        return as_dict

    @classmethod
    def from_dict(cls: Type[A], kvs: Json, *, infer_missing=False) -> A:
        doc = _decode_dataclass(cls, kvs, infer_missing)
        return doc


@dataclass
class BaseIngestDoc(ABC):
    processor_config: ProcessorConfig
    read_config: ReadConfig
    connector_config: BaseConnectorConfig

    @property
    @abstractmethod
    def unique_id(self) -> str:
        pass


@dataclass
class BaseSingleIngestDoc(BaseIngestDoc, IngestDocJsonMixin, ABC):
    """An "ingest document" is specific to a connector, and provides
    methods to fetch a single raw document, store it locally for processing, any cleanup
    needed after successful processing of the doc, and the ability to write the doc's
    structured outputs once processed.

    Crucially, it is not responsible for the actual processing of the raw document.
    """

    _source_metadata: Optional[SourceMetadata] = field(init=False, default=None)
    _date_processed: Optional[str] = field(init=False, default=None)

    @property
    def source_metadata(self) -> SourceMetadata:
        if self._source_metadata is None:
            self.update_source_metadata()
        # Provide guarantee that the field was set by update_source_metadata()
        if self._source_metadata is None:
            raise ValueError("failed to set source metadata")
        return self._source_metadata

    @source_metadata.setter
    def source_metadata(self, value: SourceMetadata):
        self._source_metadata = value

    @property
    def date_created(self) -> Optional[str]:
        """The date the document was created on the source system."""
        return self.source_metadata.date_created

    @property
    def date_modified(self) -> Optional[str]:
        """The date the document was last modified on the source system."""
        return self.source_metadata.date_modified

    @property
    def date_processed(self) -> Optional[str]:
        """The date the document was last processed by Unstructured.
        self._date_processed is assigned internally in self.partition_file()"""
        return self._date_processed

    @property
    def exists(self) -> Optional[bool]:
        """Whether the document exists on the remote source."""
        return self.source_metadata.exists

    @property
    @abstractmethod
    def filename(self):
        """The local filename of the document after fetching from remote source."""

    @property
    def base_filename(self) -> Optional[str]:
        if self.read_config.download_dir and self.filename:
            download_path = str(Path(self.read_config.download_dir).resolve())
            full_path = str(self.filename)
            base_path = full_path.replace(download_path, "")
            return base_path
        return None

    @property
    def base_output_filename(self) -> Optional[str]:
        if self.processor_config.output_dir and self._output_filename:
            output_path = str(Path(self.processor_config.output_dir).resolve())
            full_path = str(self._output_filename)
            base_path = full_path.replace(output_path, "")
            return base_path
        return None

    @property
    @abstractmethod
    def _output_filename(self):
        """Filename of the structured output for this doc."""

    @property
    def record_locator(self) -> Optional[dict[str, Any]]:  # Values must be JSON-serializable
        """A dictionary with any data necessary to uniquely identify the document on
        the source system."""
        return None

    @property
    def unique_id(self) -> str:
        return self.filename

    @property
    def source_url(self) -> Optional[str]:
        """The url of the source document."""
        return self.source_metadata.source_url  # type: ignore

    @property
    def version(self) -> Optional[str]:
        """The version of the source document, this could be the last modified date, an
        explicit version number, or anything else that can be used to uniquely identify
        the version of the document."""
        return self.source_metadata.version  # type: ignore

    @property
    def permissions_data(self) -> Optional[list[dict[str, Any]]]:
        """Access control data, aka permissions or sharing, from the source system."""
        if self.source_metadata is None:
            self.update_source_metadata()
        return self.source_metadata.permissions_data  # type: ignore

    @abstractmethod
    def cleanup_file(self):
        """Removes the local copy the file (or anything else) after successful processing."""

    @staticmethod
    def skip_if_file_exists(func):
        """Decorator that checks if a file exists, is not empty, and should not re-download,
        if so log a message indicating as much and skip the decorated function."""

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if (
                not self.read_config.re_download
                and self.filename.is_file()
                and self.filename.stat().st_size
            ):
                logger.debug(f"file exists: {self.filename}, skipping {func.__name__}")
                return None
            return func(self, *args, **kwargs)

        return wrapper

    # TODO: set as @abstractmethod and pass or raise NotImplementedError
    def update_source_metadata(self, **kwargs) -> None:
        """Sets the SourceMetadata and the  properties for the doc"""
        self._source_metadata = SourceMetadata()

    def update_permissions_data(self):
        """Sets the _permissions_data property for the doc.
        This property is later used to fill the corresponding SourceMetadata.permissions_data field,
        and after that carries on to the permissions_data property."""
        self._permissions_data: Optional[list[dict[str, Any]]] = None

    # NOTE(crag): Future BaseIngestDoc classes could define get_file_object() methods
    # in addition to or instead of get_file()
    @abstractmethod
    @SourceConnectionError.wrap
    def get_file(self):
        """Fetches the "remote" doc and stores it locally on the filesystem."""

    def has_output(self) -> bool:
        """Determine if structured output for this doc already exists."""
        return self._output_filename.is_file() and self._output_filename.stat().st_size

    @PartitionError.wrap
    def partition_file(
        self,
        partition_config: PartitionConfig,
        **partition_kwargs,
    ) -> list["Element"]:
        from unstructured.documents.elements import DataSourceMetadata
        from unstructured.partition.auto import partition
        from unstructured.staging.base import elements_from_dicts

        if not partition_config.partition_by_api:
            logger.debug("Using local partition")
            elements = partition(
                filename=str(self.filename),
                data_source_metadata=DataSourceMetadata(
                    url=self.source_url,
                    version=self.version,
                    record_locator=self.record_locator,
                    date_created=self.date_created,
                    date_modified=self.date_modified,
                    date_processed=self.date_processed,
                    permissions_data=self.permissions_data,
                ),
                **partition_kwargs,
            )
        else:
            endpoint = partition_config.partition_endpoint

            logger.debug(f"using remote partition ({endpoint})")
            elements_dicts = call_api(
                server_url=endpoint,
                api_key=partition_config.api_key,
                filename=Path(self.filename),
                api_parameters=partition_kwargs,
            )
            elements = elements_from_dicts(elements_dicts)
            # TODO: add m_data_source_metadata to unstructured-api pipeline_api and then
            # pass the stringified json here
        return elements

    def process_file(
        self,
        partition_config: PartitionConfig,
        **partition_kwargs,
    ) -> Optional[list[dict[str, Any]]]:
        self._date_processed = datetime.utcnow().isoformat()
        if self.read_config.download_only:
            return None
        logger.info(f"processing {self.filename}")

        elements = self.partition_file(partition_config=partition_config, **partition_kwargs)
        element_dicts = [e.to_dict() for e in elements]

        self.isd_elems_no_filename: list[dict[str, Any]] = []
        for elem in element_dicts:
            if partition_config.metadata_exclude and partition_config.metadata_include:
                raise ValueError(
                    "Arguments `--metadata-include` and `--metadata-exclude` are "
                    "mutually exclusive with each other.",
                )
            elif partition_config.metadata_exclude:
                ex_list = partition_config.metadata_exclude
                for ex in ex_list:
                    if "." in ex:  # handle nested fields
                        nested_fields = ex.split(".")
                        current_elem = elem
                        for f in nested_fields[:-1]:
                            if f in current_elem:
                                current_elem = current_elem[f]
                        field_to_exclude = nested_fields[-1]
                        if field_to_exclude in current_elem:
                            current_elem.pop(field_to_exclude, None)
                    else:  # handle top-level fields
                        elem["metadata"].pop(ex, None)  # type: ignore[attr-defined]
            elif partition_config.metadata_include:
                in_list = partition_config.metadata_include
                for k in list(elem["metadata"].keys()):  # type: ignore[attr-defined]
                    if k not in in_list:
                        elem["metadata"].pop(k, None)  # type: ignore[attr-defined]
            in_list = partition_config.fields_include
            elem = {k: v for k, v in elem.items() if k in in_list}

            if partition_config.flatten_metadata and "metadata" in elem:
                metadata = elem.pop("metadata")
                elem.update(flatten_dict(metadata, keys_to_omit=["data_source_record_locator"]))

            self.isd_elems_no_filename.append(elem)

        return self.isd_elems_no_filename


@dataclass
class BaseIngestDocBatch(BaseIngestDoc, BatchIngestDocJsonMixin, ABC):
    ingest_docs: list[BaseSingleIngestDoc] = field(default_factory=list)

    @abstractmethod
    @SourceConnectionError.wrap
    def get_files(self):
        """Fetches the "remote" docs and stores it locally on the filesystem."""


@dataclass
class BaseConnector(EnhancedDataClassJsonMixin, ABC):
    @abstractmethod
    def check_connection(self):
        pass


@dataclass
class BaseSourceConnector(BaseConnector, ABC):
    """Abstract Base Class for a connector to a remote source, e.g. S3 or Google Drive."""

    processor_config: ProcessorConfig
    read_config: ReadConfig
    connector_config: BaseConnectorConfig

    @abstractmethod
    def cleanup(self, cur_dir=None):
        """Any additional cleanup up need after processing is complete. E.g., removing
        temporary download dirs that are empty.

        By convention, documents that failed to process are typically not cleaned up."""

    @abstractmethod
    def initialize(self):
        """Initializes the connector. Should also validate the connector is properly
        configured: e.g., list a single a document from the source."""

    @abstractmethod
    def get_ingest_docs(self):
        """Returns all ingest docs (derived from BaseIngestDoc).
        This does not imply downloading all the raw documents themselves,
        rather each IngestDoc is capable of fetching its content (in another process)
        with IngestDoc.get_file()."""


@dataclass
class BaseDestinationConnector(BaseConnector, ABC):
    write_config: WriteConfig
    connector_config: BaseConnectorConfig

    def __init__(self, write_config: WriteConfig, connector_config: BaseConnectorConfig):
        self.write_config = write_config
        self.connector_config = connector_config

    def conform_dict(self, data: dict[str, Any]) -> None:
        """
        When the original dictionary needs to be modified in place
        """
        return

    def normalize_dict(self, element_dict: dict[str, Any]) -> dict[str, Any]:
        """
        When the original dictionary needs to be mapped to a new one
        """
        return element_dict

    @abstractmethod
    def initialize(self):
        """Initializes the connector. Should also validate the connector is properly
        configured."""

    def write(self, docs: list[BaseSingleIngestDoc]) -> None:
        elements_dict = self.get_elements_dict(docs=docs)
        self.modify_and_write_dict(elements_dict=elements_dict)

    def get_elements_dict(self, docs: list[BaseSingleIngestDoc]) -> list[dict[str, Any]]:
        dict_list: list[dict[str, Any]] = []
        for doc in docs:
            local_path = doc._output_filename
            with open(local_path) as json_file:
                dict_content = json.load(json_file)
                logger.info(
                    f"Extending {len(dict_content)} json elements from content in {local_path}",
                )
                dict_list.extend(dict_content)
        return dict_list

    @abstractmethod
    def write_dict(self, *args, elements_dict: list[dict[str, Any]], **kwargs) -> None:
        pass

    def modify_and_write_dict(self, *args, elements_dict: list[dict[str, Any]], **kwargs) -> None:
        """
        Modify in this instance means this method wraps calls to conform_dict() and
        normalize() before actually processing the content via write_dict()
        """
        for d in elements_dict:
            self.conform_dict(data=d)
        elements_dict_normalized = [self.normalize_dict(element_dict=d) for d in elements_dict]
        return self.write_dict(*args, elements_dict=elements_dict_normalized, **kwargs)

    def write_elements(self, elements: list["Element"], *args, **kwargs) -> None:
        elements_dict = [e.to_dict() for e in elements]
        self.modify_and_write_dict(*args, elements_dict=elements_dict, **kwargs)


class SourceConnectorCleanupMixin:
    read_config: ReadConfig

    def cleanup(self, cur_dir=None):
        """Recursively clean up downloaded files and directories."""
        if self.read_config.preserve_downloads or self.read_config.download_only:
            return
        if cur_dir is None:
            cur_dir = self.read_config.download_dir
        if cur_dir is None or not Path(cur_dir).is_dir():
            return
        sub_dirs = os.listdir(cur_dir)
        os.chdir(cur_dir)
        for sub_dir in sub_dirs:
            # don't traverse symlinks, not that there every should be any
            if os.path.isdir(sub_dir) and not os.path.islink(sub_dir):
                self.cleanup(sub_dir)
        os.chdir("..")
        if len(os.listdir(cur_dir)) == 0:
            os.rmdir(cur_dir)


class PermissionsCleanupMixin:
    processor_config: ProcessorConfig

    def cleanup_permissions(self, cur_dir=None):
        def has_no_folders(folder_path):
            folders = [
                item
                for item in os.listdir(folder_path)
                if os.path.isdir(os.path.join(folder_path, item))
            ]
            return len(folders) == 0

        """Recursively clean up downloaded files and directories."""
        if cur_dir is None:
            cur_dir = Path(self.processor_config.output_dir, "permissions_data")
        if not Path(cur_dir).exists():
            return
        if Path(cur_dir).is_file():
            cur_file = cur_dir
            os.remove(cur_file)
            return
        sub_dirs = os.listdir(cur_dir)
        os.chdir(cur_dir)
        for sub_dir in sub_dirs:
            # don't traverse symlinks, not that there every should be any
            if not os.path.islink(sub_dir):
                self.cleanup_permissions(sub_dir)
        os.chdir("..")
        if has_no_folders(cur_dir):
            os.rmdir(cur_dir)


class IngestDocCleanupMixin:
    read_config: ReadConfig

    @property
    @abstractmethod
    def filename(self):
        """The local filename of the document after fetching from remote source."""

    def cleanup_file(self):
        """Removes the local copy of the file after successful processing."""
        if (
            not self.read_config.preserve_downloads
            and self.filename.is_file()
            and not self.read_config.download_only
        ):
            logger.debug(f"cleaning up {self}")
            os.unlink(self.filename)


class ConfigSessionHandleMixin:
    @abstractmethod
    def create_session_handle(self) -> BaseSessionHandle:
        """Creates a session handle that will be assigned on each IngestDoc to share
        session related resources across all document handling for a given subprocess."""


@dataclass
class IngestDocSessionHandleMixin:
    connector_config: ConfigSessionHandleMixin
    _session_handle: Optional[BaseSessionHandle] = field(default=None, init=False)

    @property
    def session_handle(self):
        """If a session handle is not assigned, creates a new one and assigns it."""
        if self._session_handle is None:
            self._session_handle = self.connector_config.create_session_handle()
        return self._session_handle

    @session_handle.setter
    def session_handle(self, session_handle: BaseSessionHandle):
        self._session_handle = session_handle



import os
from dataclasses import dataclass
from typing import TYPE_CHECKING

from pydantic import Field, SecretStr

from unstructured_ingest.embed.interfaces import (
    AsyncBaseEmbeddingEncoder,
    BaseEmbeddingEncoder,
    EmbeddingConfig,
)
from unstructured_ingest.utils.dep_check import requires_dependencies

USER_AGENT = "@mixedbread-ai/unstructured"
TIMEOUT = 60
MAX_RETRIES = 3
ENCODING_FORMAT = "float"
TRUNCATION_STRATEGY = "end"


if TYPE_CHECKING:
    from mixedbread_ai.client import AsyncMixedbreadAI, MixedbreadAI
    from mixedbread_ai.core import RequestOptions


class MixedbreadAIEmbeddingConfig(EmbeddingConfig):
    """
    Configuration class for Mixedbread AI Embedding Encoder.

    Attributes:
        api_key (str): API key for accessing Mixedbread AI..
        embedder_model_name (str): Name of the model to use for embeddings.
    """

    api_key: SecretStr = Field(
        default_factory=lambda: SecretStr(os.environ.get("MXBAI_API_KEY")),
    )

    embedder_model_name: str = Field(
        default="mixedbread-ai/mxbai-embed-large-v1", alias="model_name"
    )

    @requires_dependencies(
        ["mixedbread_ai"],
        extras="embed-mixedbreadai",
    )
    def get_client(self) -> "MixedbreadAI":
        """
        Create the Mixedbread AI client.

        Returns:
            MixedbreadAI: Initialized client.
        """
        from mixedbread_ai.client import MixedbreadAI

        return MixedbreadAI(
            api_key=self.api_key.get_secret_value(),
        )

    @requires_dependencies(
        ["mixedbread_ai"],
        extras="embed-mixedbreadai",
    )
    def get_async_client(self) -> "AsyncMixedbreadAI":
        from mixedbread_ai.client import AsyncMixedbreadAI

        return AsyncMixedbreadAI(
            api_key=self.api_key.get_secret_value(),
        )


@dataclass
class MixedbreadAIEmbeddingEncoder(BaseEmbeddingEncoder):
    """
    Embedding encoder for Mixedbread AI.

    Attributes:
        config (MixedbreadAIEmbeddingConfig): Configuration for the embedding encoder.
    """

    config: MixedbreadAIEmbeddingConfig

    def get_exemplary_embedding(self) -> list[float]:
        """Get an exemplary embedding to determine dimensions and unit vector status."""
        return self.embed_query(query="Q")

    @requires_dependencies(
        ["mixedbread_ai"],
        extras="embed-mixedbreadai",
    )
    def get_request_options(self) -> "RequestOptions":
        from mixedbread_ai.core import RequestOptions

        return RequestOptions(
            max_retries=MAX_RETRIES,
            timeout_in_seconds=TIMEOUT,
            additional_headers={"User-Agent": USER_AGENT},
        )

    def get_client(self) -> "MixedbreadAI":
        return self.config.get_client()

    def embed_batch(self, client: "MixedbreadAI", batch: list[str]) -> list[list[float]]:
        response = client.embeddings(
            model=self.config.embedder_model_name,
            normalized=True,
            encoding_format=ENCODING_FORMAT,
            truncation_strategy=TRUNCATION_STRATEGY,
            request_options=self.get_request_options(),
            input=batch,
        )
        return [datum.embedding for datum in response.data]


@dataclass
class AsyncMixedbreadAIEmbeddingEncoder(AsyncBaseEmbeddingEncoder):

    config: MixedbreadAIEmbeddingConfig

    async def get_exemplary_embedding(self) -> list[float]:
        """Get an exemplary embedding to determine dimensions and unit vector status."""
        return await self.embed_query(query="Q")

    @requires_dependencies(
        ["mixedbread_ai"],
        extras="embed-mixedbreadai",
    )
    def get_request_options(self) -> "RequestOptions":
        from mixedbread_ai.core import RequestOptions

        return RequestOptions(
            max_retries=MAX_RETRIES,
            timeout_in_seconds=TIMEOUT,
            additional_headers={"User-Agent": USER_AGENT},
        )

    def get_client(self) -> "AsyncMixedbreadAI":
        return self.config.get_async_client()

    async def embed_batch(self, client: "AsyncMixedbreadAI", batch: list[str]) -> list[list[float]]:
        response = await client.embeddings(
            model=self.config.embedder_model_name,
            normalized=True,
            encoding_format=ENCODING_FORMAT,
            truncation_strategy=TRUNCATION_STRATEGY,
            request_options=self.get_request_options(),
            input=batch,
        )
        return [datum.embedding for datum in response.data]



import os
import typing as t
from dataclasses import dataclass
from datetime import datetime as dt
from multiprocessing import Process
from pathlib import Path

from unstructured_ingest.error import SourceConnectionError, SourceConnectionNetworkError
from unstructured_ingest.interfaces import (
    BaseConnectorConfig,
    BaseDestinationConnector,
    BaseSingleIngestDoc,
    BaseSourceConnector,
    IngestDocCleanupMixin,
    SourceConnectorCleanupMixin,
    SourceMetadata,
    WriteConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.dep_check import requires_dependencies

if t.TYPE_CHECKING:
    from deltalake import DeltaTable


@dataclass
class SimpleDeltaTableConfig(BaseConnectorConfig):
    table_uri: t.Union[str, Path]
    version: t.Optional[int] = None
    storage_options: t.Optional[t.Dict[str, str]] = None
    without_files: bool = False


@dataclass
class DeltaTableIngestDoc(IngestDocCleanupMixin, BaseSingleIngestDoc):
    connector_config: SimpleDeltaTableConfig
    uri: str
    modified_date: str
    created_at: str
    registry_name: str = "delta-table"

    def uri_filename(self) -> str:
        basename = os.path.basename(self.uri)
        return os.path.splitext(basename)[0]

    @property
    def filename(self):
        return (Path(self.read_config.download_dir) / f"{self.uri_filename()}.csv").resolve()

    @property
    def _output_filename(self):
        """Create filename document id combined with a hash of the query to uniquely identify
        the output file."""
        return Path(self.processor_config.output_dir) / f"{self.uri_filename()}.json"

    def _create_full_tmp_dir_path(self):
        self.filename.parent.mkdir(parents=True, exist_ok=True)
        self._output_filename.parent.mkdir(parents=True, exist_ok=True)

    @requires_dependencies(["fsspec"], extras="delta-table")
    def _get_fs_from_uri(self):
        from fsspec.core import url_to_fs

        try:
            fs, _ = url_to_fs(self.uri)
        except ImportError as error:
            raise ImportError(
                f"uri {self.uri} may be associated with a filesystem that "
                f"requires additional dependencies: {error}",
            )
        return fs

    def update_source_metadata(self, **kwargs):
        fs = kwargs.get("fs", self._get_fs_from_uri())
        version = (
            fs.checksum(self.uri) if fs.protocol != "gs" else fs.info(self.uri).get("etag", "")
        )
        file_exists = fs.exists(self.uri)
        self.source_metadata = SourceMetadata(
            date_created=self.created_at,
            date_modified=self.modified_date,
            version=version,
            source_url=self.uri,
            exists=file_exists,
        )

    @SourceConnectionError.wrap
    @BaseSingleIngestDoc.skip_if_file_exists
    def get_file(self):
        fs = self._get_fs_from_uri()
        self.update_source_metadata(fs=fs)
        logger.info(f"using a {fs} filesystem to collect table data")
        self._create_full_tmp_dir_path()

        df = self._get_df(filesystem=fs)

        logger.info(f"writing {len(df)} rows to {self.filename}")
        df.to_csv(self.filename)

    @SourceConnectionNetworkError.wrap
    def _get_df(self, filesystem):
        import pyarrow.parquet as pq

        return pq.ParquetDataset(self.uri, filesystem=filesystem).read_pandas().to_pandas()


@dataclass
class DeltaTableSourceConnector(SourceConnectorCleanupMixin, BaseSourceConnector):
    connector_config: SimpleDeltaTableConfig
    delta_table: t.Optional["DeltaTable"] = None

    def check_connection(self):
        pass

    @requires_dependencies(["deltalake"], extras="delta-table")
    def initialize(self):
        from deltalake import DeltaTable

        self.delta_table = DeltaTable(
            table_uri=self.connector_config.table_uri,
            version=self.connector_config.version,
            storage_options=self.connector_config.storage_options,
            without_files=self.connector_config.without_files,
        )
        rows = self.delta_table.to_pyarrow_dataset().count_rows()
        if not rows > 0:
            raise ValueError(f"no data found at {self.connector_config.table_uri}")
        logger.info(f"processing {rows} rows of data")

    def get_ingest_docs(self):
        """Batches the results into distinct docs"""
        if not self.delta_table:
            raise ValueError("delta table was never initialized")
        actions = self.delta_table.get_add_actions().to_pandas()
        mod_date_dict = {
            row["path"]: str(row["modification_time"]) for _, row in actions.iterrows()
        }
        created_at = dt.fromtimestamp(self.delta_table.metadata().created_time / 1000)
        return [
            DeltaTableIngestDoc(
                connector_config=self.connector_config,
                processor_config=self.processor_config,
                read_config=self.read_config,
                uri=uri,
                modified_date=mod_date_dict[os.path.basename(uri)],
                created_at=str(created_at),
            )
            for uri in self.delta_table.file_uris()
        ]


@dataclass
class DeltaTableWriteConfig(WriteConfig):
    drop_empty_cols: bool = False
    mode: t.Literal["error", "append", "overwrite", "ignore"] = "error"
    schema_mode: t.Optional[t.Literal["merge", "overwrite"]] = None
    engine: t.Literal["pyarrow", "rust"] = "pyarrow"


@dataclass
class DeltaTableDestinationConnector(BaseDestinationConnector):
    write_config: DeltaTableWriteConfig
    connector_config: SimpleDeltaTableConfig

    @requires_dependencies(["deltalake"], extras="delta-table")
    def initialize(self):
        pass

    def check_connection(self):
        pass

    @requires_dependencies(["deltalake"], extras="delta-table")
    def write_dict(self, *args, elements_dict: t.List[t.Dict[str, t.Any]], **kwargs) -> None:
        from deltalake.writer import write_deltalake

        from unstructured_ingest.utils.table import convert_to_pandas_dataframe

        df = convert_to_pandas_dataframe(
            elements_dict=elements_dict,
            drop_empty_cols=self.write_config.drop_empty_cols,
        )
        logger.info(
            f"writing {len(df)} rows to destination table "
            f"at {self.connector_config.table_uri}\ndtypes: {df.dtypes}",
        )
        writer_kwargs = {
            "table_or_uri": self.connector_config.table_uri,
            "data": df,
            "mode": self.write_config.mode,
            "engine": self.write_config.engine,
        }
        if self.write_config.schema_mode is not None:
            writer_kwargs["schema_mode"] = self.write_config.schema_mode
        # NOTE: deltalake writer on Linux sometimes can finish but still trigger a SIGABRT and cause
        # ingest to fail, even though all tasks are completed normally. Putting the writer into a
        # process mitigates this issue by ensuring python interpreter waits properly for deltalake's
        # rust backend to finish
        writer = Process(
            target=write_deltalake,
            kwargs=writer_kwargs,
        )
        writer.start()
        writer.join()



from dataclasses import fields
from pathlib import Path
from typing import TYPE_CHECKING, Optional

from unstructured_ingest.v2.errors import ProviderError, QuotaError, UserAuthError, UserError
from unstructured_ingest.v2.logger import logger

if TYPE_CHECKING:
    from unstructured_client.models.operations import PartitionRequest


def create_partition_request(filename: Path, parameters_dict: dict) -> "PartitionRequest":
    """Given a filename and a dict of API parameters, return a PartitionRequest for use
    by unstructured-client. Remove any params that aren't recognized by the SDK.

    Args:
        filename: Path to the file being partitioned
        parameters_dict: A mapping of all API params we want to send

    Returns: A PartitionRequest containing the file and all valid params
    """
    from unstructured_client.models.operations import PartitionRequest
    from unstructured_client.models.shared import Files, PartitionParameters

    # NOTE(austin): PartitionParameters is a Pydantic model in v0.26.0
    # Prior to this it was a dataclass which doesn't have .__fields
    try:
        possible_fields = PartitionParameters.model_fields
    except AttributeError:
        possible_fields = [f.name for f in fields(PartitionParameters)]

    filtered_partition_request = {k: v for k, v in parameters_dict.items() if k in possible_fields}
    if len(filtered_partition_request) != len(parameters_dict):
        logger.debug(
            "Following fields were omitted due to not being "
            "supported by the currently used unstructured client: {}".format(
                ", ".join([v for v in parameters_dict if v not in filtered_partition_request])
            )
        )

    logger.debug(f"using hosted partitioner with kwargs: {parameters_dict}")

    with open(filename, "rb") as f:
        files = Files(
            content=f.read(),
            file_name=str(filename.resolve()),
        )
        filtered_partition_request["files"] = files

    partition_params = PartitionParameters(**filtered_partition_request)

    return PartitionRequest(partition_parameters=partition_params)


def wrap_error(e: Exception) -> Exception:
    from unstructured_client.models.errors.sdkerror import SDKError

    if not isinstance(e, SDKError):
        logger.error(f"Uncaught Error calling API: {e}")
        raise e
    status_code = e.status_code
    body = e.body
    if status_code == 402:
        return QuotaError(body)
    if status_code in [401, 403]:
        return UserAuthError(body)
    if 400 <= status_code < 500:
        return UserError(body)
    if status_code >= 500:
        return ProviderError(body)
    logger.error(f"Uncaught Error calling API: {e}")
    raise e


async def call_api_async(
    server_url: Optional[str], api_key: Optional[str], filename: Path, api_parameters: dict
) -> list[dict]:
    """Call the Unstructured API using unstructured-client.

    Args:
        server_url: The base URL where the API is hosted
        api_key: The user's API key (can be empty if this is a self hosted API)
        filename: Path to the file being partitioned
        api_parameters: A dict containing the requested API parameters

    Returns: A list of the file's elements, or an empty list if there was an error
    """
    from unstructured_client import UnstructuredClient

    client = UnstructuredClient(
        server_url=server_url,
        api_key_auth=api_key,
    )
    partition_request = create_partition_request(filename=filename, parameters_dict=api_parameters)
    try:
        res = await client.general.partition_async(request=partition_request)
    except Exception as e:
        raise wrap_error(e)

    return res.elements or []


def call_api(
    server_url: Optional[str], api_key: Optional[str], filename: Path, api_parameters: dict
) -> list[dict]:
    """Call the Unstructured API using unstructured-client.

    Args:
        server_url: The base URL where the API is hosted
        api_key: The user's API key (can be empty if this is a self hosted API)
        filename: Path to the file being partitioned
        api_parameters: A dict containing the requested API parameters

    Returns: A list of the file's elements, or an empty list if there was an error
    """
    from unstructured_client import UnstructuredClient

    client = UnstructuredClient(
        server_url=server_url,
        api_key_auth=api_key,
    )
    partition_request = create_partition_request(filename=filename, parameters_dict=api_parameters)
    try:
        res = client.general.partition(request=partition_request)
    except Exception as e:
        raise wrap_error(e)

    return res.elements or []



import typing as t
from dataclasses import dataclass

from unstructured_ingest.pipeline.interfaces import DocFactoryNode


@dataclass
class DocFactory(DocFactoryNode):
    def run(self, *args, **kwargs) -> t.Iterable[dict]:
        docs = self.source_doc_connector.get_ingest_docs()
        json_docs = [doc.to_dict() for doc in docs]
        return json_docs



from dataclasses import dataclass
from pathlib import Path
from typing import Any

import pandas as pd

from unstructured_ingest.utils.data_prep import get_data, write_data
from unstructured_ingest.v2.interfaces import FileData, UploadStager
from unstructured_ingest.v2.utils import get_enhanced_element_id

_COLUMNS = (
    "id",
    "element_id",
    "text",
    "embeddings",
    "type",
    "system",
    "layout_width",
    "layout_height",
    "points",
    "url",
    "version",
    "date_created",
    "date_modified",
    "date_processed",
    "permissions_data",
    "record_locator",
    "category_depth",
    "parent_id",
    "attached_filename",
    "filetype",
    "last_modified",
    "file_directory",
    "filename",
    "languages",
    "page_number",
    "links",
    "page_name",
    "link_urls",
    "link_texts",
    "sent_from",
    "sent_to",
    "subject",
    "section",
    "header_footer_type",
    "emphasized_text_contents",
    "emphasized_text_tags",
    "text_as_html",
    "regex_metadata",
    "detection_class_prob",
)

# _DATE_COLUMNS = ("date_created", "date_modified", "date_processed", "last_modified")


@dataclass
class BaseDuckDBUploadStager(UploadStager):

    def conform_dict(self, element_dict: dict, file_data: FileData) -> dict:
        data = element_dict.copy()
        metadata: dict[str, Any] = data.pop("metadata", {})
        data_source = metadata.pop("data_source", {})
        coordinates = metadata.pop("coordinates", {})

        data.update(metadata)
        data.update(data_source)
        data.update(coordinates)

        data["id"] = get_enhanced_element_id(element_dict=data, file_data=file_data)

        # remove extraneous, not supported columns
        data = {k: v for k, v in data.items() if k in _COLUMNS}
        return data

    def run(
        self,
        elements_filepath: Path,
        file_data: FileData,
        output_dir: Path,
        output_filename: str,
        **kwargs: Any,
    ) -> Path:
        elements_contents = get_data(path=elements_filepath)
        output_filename_suffix = Path(elements_filepath).suffix
        output_filename = f"{Path(output_filename).stem}{output_filename_suffix}"
        output_path = self.get_output_path(output_filename=output_filename, output_dir=output_dir)

        output = [
            self.conform_dict(element_dict=element_dict, file_data=file_data)
            for element_dict in elements_contents
        ]
        df = pd.DataFrame(data=output)

        for column in filter(
            lambda x: x in df.columns,
            ("version", "page_number", "regex_metadata"),
        ):
            df[column] = df[column].apply(str)

        data = df.to_dict(orient="records")
        write_data(path=output_path, data=data)
        return output_path



import typing as t
from dataclasses import dataclass

import click

from unstructured_ingest.cli.base.src import BaseSrcCmd
from unstructured_ingest.cli.interfaces import (
    CliConfig,
    CliRecursiveConfig,
    DelimitedString,
)
from unstructured_ingest.connector.notion.connector import SimpleNotionConfig


@dataclass
class NotionCliConfig(SimpleNotionConfig, CliConfig):
    @staticmethod
    def get_cli_options() -> t.List[click.Option]:
        options = [
            click.Option(
                ["--notion-api-key"],
                required=True,
                type=str,
                help="API key for Notion api",
            ),
            click.Option(
                ["--page-ids"],
                default=None,
                type=DelimitedString(),
                help="Notion page IDs to pull text from",
            ),
            click.Option(
                ["--database-ids"],
                default=None,
                type=DelimitedString(),
                help="Notion database IDs to pull text from",
            ),
        ]
        return options


def get_base_src_cmd() -> BaseSrcCmd:
    cmd_cls = BaseSrcCmd(
        cmd_name="notion",
        cli_config=NotionCliConfig,
        additional_cli_options=[CliRecursiveConfig],
    )
    return cmd_cls



# https://developers.notion.com/reference/block#image
from typing import Optional

from htmlBuilder.attributes import Src
from htmlBuilder.tags import HtmlTag, Source
from htmlBuilder.tags import Video as VideoHtml

from unstructured_ingest.connector.notion.interfaces import BlockBase
from unstructured_ingest.connector.notion.types.file import FileObject


class Video(BlockBase, FileObject):
    @staticmethod
    def can_have_children() -> bool:
        return False

    def get_html(self) -> Optional[HtmlTag]:
        if self.external:
            return VideoHtml([], [Source([Src(self.external.url)], [self.external.url])])
        if self.file:
            return VideoHtml([], [Source([Src(self.file.url)], [self.file.url])])
        return None



import json
from pathlib import Path

import chromadb
import pytest
from _pytest.fixtures import TopRequest

from test.integration.connectors.utils.constants import DESTINATION_TAG, VECTOR_DB_TAG
from test.integration.connectors.utils.docker import HealthCheck, container_context
from test.integration.connectors.utils.validation.destination import (
    StagerValidationConfigs,
    stager_validation,
)
from unstructured_ingest.v2.interfaces import FileData, SourceIdentifiers
from unstructured_ingest.v2.processes.connectors.chroma import (
    CONNECTOR_TYPE,
    ChromaConnectionConfig,
    ChromaUploader,
    ChromaUploaderConfig,
    ChromaUploadStager,
    ChromaUploadStagerConfig,
)


@pytest.fixture
def chroma_instance():
    with container_context(
        image="chromadb/chroma:0.6.2",
        ports={8000: 8000},
        name="chroma_int_test",
        healthcheck=HealthCheck(
            interval=5,
            timeout=10,
            retries=3,
            test="timeout 10s bash -c ':> /dev/tcp/127.0.0.1/8000' || exit 1",
        ),
    ) as ctx:
        yield ctx


def validate_collection(collection_name: str, num_embeddings: int):
    print(f"Checking contents of Chroma collection: {collection_name}")

    chroma_client = chromadb.HttpClient(
        host="localhost",
        port="8000",
        tenant="default_tenant",
        database="default_database",
    )

    collection = chroma_client.get_or_create_collection(name=collection_name)

    number_of_embeddings = collection.count()
    expected_embeddings = num_embeddings
    print(
        f"# of embeddings in collection vs expected: {number_of_embeddings}/{expected_embeddings}"
    )

    assert number_of_embeddings == expected_embeddings, (
        f"Number of rows in generated table ({number_of_embeddings}) "
        f"doesn't match expected value: {expected_embeddings}"
    )


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
def test_chroma_destination(
    upload_file: Path,
    chroma_instance,
    tmp_path: Path,
):
    collection_name = "test_collection"
    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=upload_file.name, filename=upload_file.name),
        connector_type=CONNECTOR_TYPE,
        identifier="mock file data",
    )
    stager = ChromaUploadStager(upload_stager_config=ChromaUploadStagerConfig())

    uploader = ChromaUploader(
        connection_config=ChromaConnectionConfig(
            host="localhost",
            port=8000,
            tenant="default_tenant",
            database="default_database",
        ),
        upload_config=ChromaUploaderConfig(collection_name=collection_name),
    )
    staged_filepath = stager.run(
        elements_filepath=upload_file,
        file_data=file_data,
        output_dir=tmp_path,
        output_filename=upload_file.name,
    )
    uploader.precheck()
    uploader.run(path=staged_filepath, file_data=file_data)

    # Run validation
    with staged_filepath.open() as f:
        staged_elements = json.load(f)
    expected_count = len(staged_elements)
    validate_collection(collection_name=collection_name, num_embeddings=expected_count)


@pytest.mark.parametrize("upload_file_str", ["upload_file_ndjson", "upload_file"])
@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, "stager", VECTOR_DB_TAG)
def test_chroma_stager(
    request: TopRequest,
    upload_file_str: str,
    tmp_path: Path,
):
    upload_file: Path = request.getfixturevalue(upload_file_str)
    stager = ChromaUploadStager()
    stager_validation(
        configs=StagerValidationConfigs(test_id=CONNECTOR_TYPE, expected_count=22),
        input_file=upload_file,
        stager=stager,
        tmp_dir=tmp_path,
    )



class UserError(Exception):
    pass


class UserAuthError(UserError):
    pass


class RateLimitError(UserError):
    pass


class QuotaError(UserError):
    pass


class ProviderError(Exception):
    pass


recognized_errors = [UserError, UserAuthError, RateLimitError, QuotaError, ProviderError]


def is_internal_error(e: Exception) -> bool:
    return any(isinstance(e, recognized_error) for recognized_error in recognized_errors)



import logging

from unstructured_ingest.logger import ingest_log_streaming_init


def log_options(options: dict, verbose=False):
    ingest_log_streaming_init(logging.DEBUG if verbose else logging.INFO)



import typing as t
from dataclasses import dataclass

import click

from unstructured_ingest.cli.interfaces import CliConfig
from unstructured_ingest.connector.databricks_volumes import (
    DatabricksVolumesWriteConfig,
    SimpleDatabricksVolumesConfig,
)

CMD_NAME = "databricks-volumes"


@dataclass
class DatabricksVolumesCliConfig(SimpleDatabricksVolumesConfig, CliConfig):
    @staticmethod
    def get_cli_options() -> t.List[click.Option]:
        options = [
            click.Option(
                ["--host"],
                type=str,
                default=None,
                help="The Databricks host URL for either the "
                "Databricks workspace endpoint or the "
                "Databricks accounts endpoint.",
            ),
            click.Option(
                ["--account-id"],
                type=str,
                default=None,
                help="The Databricks account ID for the Databricks "
                "accounts endpoint. Only has effect when Host is "
                "either https://accounts.cloud.databricks.com/ (AWS), "
                "https://accounts.azuredatabricks.net/ (Azure), "
                "or https://accounts.gcp.databricks.com/ (GCP).",
            ),
            click.Option(
                ["--username"],
                type=str,
                default=None,
                help="The Databricks username part of basic authentication. "
                "Only possible when Host is *.cloud.databricks.com (AWS).",
            ),
            click.Option(
                ["--password"],
                type=str,
                default=None,
                help="The Databricks password part of basic authentication. "
                "Only possible when Host is *.cloud.databricks.com (AWS).",
            ),
            click.Option(["--client-id"], type=str, default=None),
            click.Option(["--client-secret"], type=str, default=None),
            click.Option(
                ["--token"],
                type=str,
                default=None,
                help="The Databricks personal access token (PAT) (AWS, Azure, and GCP) or "
                "Azure Active Directory (Azure AD) token (Azure).",
            ),
            click.Option(
                ["--azure-workspace-resource-id"],
                type=str,
                default=None,
                help="The Azure Resource Manager ID for the Azure Databricks workspace, "
                "which is exchanged for a Databricks host URL.",
            ),
            click.Option(
                ["--azure-client-secret"],
                type=str,
                default=None,
                help="The Azure AD service principal’s client secret.",
            ),
            click.Option(
                ["--azure-client-id"],
                type=str,
                default=None,
                help="The Azure AD service principal’s application ID.",
            ),
            click.Option(
                ["--azure-tenant-id"],
                type=str,
                default=None,
                help="The Azure AD service principal’s tenant ID.",
            ),
            click.Option(
                ["--azure-environment"],
                type=str,
                default=None,
                help="The Azure environment type (such as Public, UsGov, China, and Germany) for a "
                "specific set of API endpoints. Defaults to PUBLIC.",
            ),
            click.Option(
                ["--auth-type"],
                type=str,
                default=None,
                help="When multiple auth attributes are available in the "
                "environment, use the auth type specified by this "
                "argument. This argument also holds the currently "
                "selected auth.",
            ),
            click.Option(["--cluster-id"], type=str, default=None),
            click.Option(["--google-credentials"], type=str, default=None),
            click.Option(["--google-service-account"], type=str, default=None),
        ]
        return options


@dataclass
class DatabricksVolumesCliWriteConfig(DatabricksVolumesWriteConfig, CliConfig):
    @staticmethod
    def get_cli_options() -> t.List[click.Option]:
        options = [
            click.Option(
                ["--volume"], type=str, required=True, help="Name of volume in the Unity Catalog"
            ),
            click.Option(
                ["--catalog"],
                type=str,
                required=True,
                help="Name of the catalog in the Databricks Unity Catalog service",
            ),
            click.Option(
                ["--volume-path"],
                type=str,
                required=False,
                default=None,
                help="Optional path within the volume to write to",
            ),
            click.Option(
                ["--overwrite"],
                type=bool,
                is_flag=True,
                help="If true, an existing file will be overwritten.",
            ),
            click.Option(
                ["--encoding"],
                type=str,
                required=True,
                default="utf-8",
                help="Encoding applied to the data when written to the volume",
            ),
            click.Option(
                ["--schema"],
                type=str,
                required=True,
                default="default",
                help="Schema associated with the volume to write to in the Unity Catalog service",
            ),
        ]
        return options


def get_base_dest_cmd():
    from unstructured_ingest.cli.base.dest import BaseDestCmd

    cmd_cls = BaseDestCmd(
        cmd_name=CMD_NAME,
        cli_config=DatabricksVolumesCliConfig,
        additional_cli_options=[DatabricksVolumesCliWriteConfig],
        write_config=DatabricksVolumesWriteConfig,
    )
    return cmd_cls



# https://developers.notion.com/reference/property-object#people
from dataclasses import dataclass, field
from typing import List, Optional

from htmlBuilder.tags import Div, HtmlTag, Span

from unstructured_ingest.connector.notion.interfaces import DBCellBase, DBPropertyBase
from unstructured_ingest.connector.notion.types.user import People as PeopleType


@dataclass
class People(DBPropertyBase):
    id: str
    name: str
    type: str = "people"
    people: dict = field(default_factory=dict)

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class PeopleCell(DBCellBase):
    id: str
    people: List[PeopleType]
    type: str = "people"
    name: Optional[str] = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(people=[PeopleType.from_dict(p) for p in data.pop("people", {})], **data)

    def get_html(self) -> Optional[HtmlTag]:
        if not self.people:
            return None
        people_spans = []
        for person in self.people:
            people_spans.append(Span([], person.get_html()))
        return Div([], people_spans)



from dataclasses import dataclass
from typing import Optional

from pydantic import Field, Secret

from unstructured_ingest.v2.processes.connector_registry import (
    DestinationRegistryEntry,
    SourceRegistryEntry,
)
from unstructured_ingest.v2.processes.connectors.databricks.volumes import (
    DatabricksVolumesAccessConfig,
    DatabricksVolumesConnectionConfig,
    DatabricksVolumesDownloader,
    DatabricksVolumesDownloaderConfig,
    DatabricksVolumesIndexer,
    DatabricksVolumesIndexerConfig,
    DatabricksVolumesUploader,
    DatabricksVolumesUploaderConfig,
)

CONNECTOR_TYPE = "databricks_volumes"


class DatabricksNativeVolumesAccessConfig(DatabricksVolumesAccessConfig):
    client_id: Optional[str] = Field(default=None, description="Client ID of the OAuth app.")
    client_secret: Optional[str] = Field(
        default=None, description="Client Secret of the OAuth app."
    )
    profile: Optional[str] = None
    azure_workspace_resource_id: Optional[str] = Field(
        default=None,
        description="The Azure Resource Manager ID for the Azure Databricks workspace, "
        "which is exchanged for a Databricks host URL.",
    )


class DatabricksNativeVolumesConnectionConfig(DatabricksVolumesConnectionConfig):
    access_config: Secret[DatabricksNativeVolumesAccessConfig]


class DatabricksNativeVolumesIndexerConfig(DatabricksVolumesIndexerConfig):
    pass


@dataclass
class DatabricksNativeVolumesIndexer(DatabricksVolumesIndexer):
    connection_config: DatabricksNativeVolumesConnectionConfig
    index_config: DatabricksNativeVolumesIndexerConfig
    connector_type: str = CONNECTOR_TYPE


class DatabricksNativeVolumesDownloaderConfig(DatabricksVolumesDownloaderConfig):
    pass


@dataclass
class DatabricksNativeVolumesDownloader(DatabricksVolumesDownloader):
    connection_config: DatabricksNativeVolumesConnectionConfig
    download_config: DatabricksVolumesDownloaderConfig
    connector_type: str = CONNECTOR_TYPE


class DatabricksNativeVolumesUploaderConfig(DatabricksVolumesUploaderConfig):
    pass


@dataclass
class DatabricksNativeVolumesUploader(DatabricksVolumesUploader):
    connection_config: DatabricksNativeVolumesConnectionConfig
    upload_config: DatabricksNativeVolumesUploaderConfig
    connector_type: str = CONNECTOR_TYPE


databricks_native_volumes_destination_entry = DestinationRegistryEntry(
    connection_config=DatabricksNativeVolumesConnectionConfig,
    uploader=DatabricksNativeVolumesUploader,
    uploader_config=DatabricksNativeVolumesUploaderConfig,
)

databricks_native_volumes_source_entry = SourceRegistryEntry(
    connection_config=DatabricksNativeVolumesConnectionConfig,
    indexer=DatabricksNativeVolumesIndexer,
    indexer_config=DatabricksNativeVolumesIndexerConfig,
    downloader=DatabricksNativeVolumesDownloader,
    downloader_config=DatabricksNativeVolumesDownloaderConfig,
)



"""
Salesforce Connector
Able to download Account, Case, Campaign, EmailMessage, Lead
Salesforce returns everything as a list of json.
This saves each entry as a separate file to be partitioned.
Using JWT authorization
https://developer.salesforce.com/docs/atlas.en-us.sfdx_dev.meta/sfdx_dev/sfdx_dev_auth_key_and_cert.htm
https://developer.salesforce.com/docs/atlas.en-us.sfdx_dev.meta/sfdx_dev/sfdx_dev_auth_connected_app.htm
"""

import json
import typing as t
from collections import OrderedDict
from dataclasses import dataclass, field
from datetime import datetime
from email.utils import formatdate
from pathlib import Path
from string import Template
from textwrap import dedent

from unstructured_ingest.enhanced_dataclass import enhanced_field
from unstructured_ingest.error import SourceConnectionError, SourceConnectionNetworkError
from unstructured_ingest.interfaces import (
    AccessConfig,
    BaseConnectorConfig,
    BaseSingleIngestDoc,
    BaseSourceConnector,
    IngestDocCleanupMixin,
    SourceConnectorCleanupMixin,
    SourceMetadata,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.dep_check import requires_dependencies


class MissingCategoryError(Exception):
    """There are no categories with that name."""


SALESFORCE_API_VERSION = "57.0"

ACCEPTED_CATEGORIES = ["Account", "Case", "Campaign", "EmailMessage", "Lead"]

EMAIL_TEMPLATE = Template(
    """MIME-Version: 1.0
Date: $date
Message-ID: $message_identifier
Subject: $subject
From: $from_email
To: $to_email
Content-Type: multipart/alternative; boundary="00000000000095c9b205eff92630"
--00000000000095c9b205eff92630
Content-Type: text/plain; charset="UTF-8"
$textbody
--00000000000095c9b205eff92630
Content-Type: text/html; charset="UTF-8"
$htmlbody
--00000000000095c9b205eff92630--
""",
)


@dataclass
class SalesforceAccessConfig(AccessConfig):
    consumer_key: str = enhanced_field(sensitive=True)
    private_key: str = enhanced_field(sensitive=True)

    @requires_dependencies(["cryptography"])
    def get_private_key_value_and_type(self) -> t.Tuple[str, t.Type]:
        from cryptography.hazmat.primitives import serialization

        try:
            serialization.load_pem_private_key(data=self.private_key.encode("utf-8"), password=None)
        except ValueError:
            pass
        else:
            return self.private_key, str

        if Path(self.private_key).is_file():
            return self.private_key, Path

        raise ValueError("private_key does not contain PEM private key or path")


@dataclass
class SimpleSalesforceConfig(BaseConnectorConfig):
    """Connector specific attributes"""

    access_config: SalesforceAccessConfig
    categories: t.List[str]
    username: str
    recursive: bool = False

    @requires_dependencies(["simple_salesforce"], extras="salesforce")
    def get_client(self):
        from simple_salesforce import Salesforce

        pkey_value, pkey_type = self.access_config.get_private_key_value_and_type()

        return Salesforce(
            username=self.username,
            consumer_key=self.access_config.consumer_key,
            privatekey_file=pkey_value if pkey_type is Path else None,
            privatekey=pkey_value if pkey_type is str else None,
            version=SALESFORCE_API_VERSION,
        )


@dataclass
class SalesforceIngestDoc(IngestDocCleanupMixin, BaseSingleIngestDoc):
    connector_config: SimpleSalesforceConfig
    record_type: str
    record_id: str
    registry_name: str = "salesforce"
    _record: OrderedDict = field(default_factory=lambda: OrderedDict())

    @property
    def record(self):
        if not self._record:
            self._record = self.get_record()
        return self._record

    def get_file_extension(self) -> str:
        if self.record_type == "EmailMessage":
            extension = ".eml"
        elif self.record_type in ["Account", "Lead", "Case", "Campaign"]:
            extension = ".xml"
        else:
            raise MissingCategoryError(
                f"There are no categories with the name: {self.record_type}",
            )
        return extension

    def _tmp_download_file(self) -> Path:
        record_file = self.record_id + self.get_file_extension()
        return Path(self.read_config.download_dir) / self.record_type / record_file

    @property
    def _output_filename(self) -> Path:
        record_file = self.record_id + self.get_file_extension() + ".json"
        return Path(self.processor_config.output_dir) / self.record_type / record_file

    def _create_full_tmp_dir_path(self):
        self._tmp_download_file().parent.mkdir(parents=True, exist_ok=True)

    def _xml_for_record(self, record: OrderedDict) -> str:
        """Creates partitionable xml file from a record"""
        import xml.etree.ElementTree as ET

        def flatten_dict(data, parent, prefix=""):
            for key, value in data.items():
                if isinstance(value, OrderedDict):
                    flatten_dict(value, parent, prefix=f"{prefix}{key}.")
                else:
                    item = ET.Element("item")
                    item.text = f"{prefix}{key}: {value}"
                    parent.append(item)

        root = ET.Element("root")
        flatten_dict(record, root)
        xml_string = ET.tostring(root, encoding="utf-8", xml_declaration=True).decode()
        return xml_string

    def _eml_for_record(self, email_json: t.Dict[str, t.Any]) -> str:
        from dateutil import parser  # type: ignore

        """Recreates standard expected .eml format using template."""
        eml = EMAIL_TEMPLATE.substitute(
            date=formatdate(parser.parse(email_json.get("MessageDate")).timestamp()),
            message_identifier=email_json.get("MessageIdentifier"),
            subject=email_json.get("Subject"),
            from_email=email_json.get("FromAddress"),
            to_email=email_json.get("ToAddress"),
            textbody=email_json.get("TextBody"),
            # TODO: This is a hack to get emails to process correctly.
            # The HTML partitioner seems to have issues with <br> and text without tags like <p>
            htmlbody=email_json.get("HtmlBody", "")  # "" because you can't .replace None
            .replace("<br />", "<p>")
            .replace("<body", "<body><p"),
        )
        return dedent(eml)

    @SourceConnectionNetworkError.wrap
    def _get_response(self):
        client = self.connector_config.get_client()
        return client.query_all(
            f"select FIELDS(STANDARD) from {self.record_type} where Id='{self.record_id}'",
        )

    def get_record(self) -> OrderedDict:
        # Get record from Salesforce based on id
        response = self._get_response()
        logger.debug(f"response was returned for salesforce record id: {self.record_id}")
        records = response["records"]
        if not records:
            raise ValueError(
                f"No record found with record id {self.record_id}: {json.dumps(response)}"
            )
        record_json = records[0]
        return record_json

    def update_source_metadata(self) -> None:  # type: ignore
        record_json = self.record

        date_format = "%Y-%m-%dT%H:%M:%S.000+0000"
        self.source_metadata = SourceMetadata(
            date_created=datetime.strptime(record_json["CreatedDate"], date_format).isoformat(),
            date_modified=datetime.strptime(
                record_json["LastModifiedDate"],
                date_format,
            ).isoformat(),
            # SystemModstamp is Timestamp if record has been modified by person or automated system
            version=record_json.get("SystemModstamp"),
            source_url=record_json["attributes"].get("url"),
            exists=True,
        )

    @SourceConnectionError.wrap
    @BaseSingleIngestDoc.skip_if_file_exists
    def get_file(self):
        """Saves individual json records locally."""
        self._create_full_tmp_dir_path()
        record = self.record

        self.update_source_metadata()

        try:
            if self.record_type == "EmailMessage":
                document = self._eml_for_record(record)
            else:
                document = self._xml_for_record(record)

            with open(self._tmp_download_file(), "w") as page_file:
                page_file.write(document)

        except Exception as e:
            logger.error(
                f"Error while downloading and saving file: {self.record_id}.",
            )
            logger.error(e)

    @property
    def filename(self):
        """The filename of the file created from a Salesforce record"""
        return self._tmp_download_file()


@dataclass
class SalesforceSourceConnector(SourceConnectorCleanupMixin, BaseSourceConnector):
    connector_config: SimpleSalesforceConfig

    def __post_init__(self):
        self.ingest_doc_cls: t.Type[SalesforceIngestDoc] = SalesforceIngestDoc

    def initialize(self):
        pass

    @requires_dependencies(["simple_salesforce"], extras="salesforce")
    def check_connection(self):
        from simple_salesforce.exceptions import SalesforceError

        try:
            self.connector_config.get_client()
        except SalesforceError as salesforce_error:
            logger.error(f"failed to validate connection: {salesforce_error}", exc_info=True)
            raise SourceConnectionError(f"failed to validate connection: {salesforce_error}")

    @requires_dependencies(["simple_salesforce"], extras="salesforce")
    def get_ingest_docs(self) -> t.List[SalesforceIngestDoc]:
        """Get Salesforce Ids for the records.
        Send them to next phase where each doc gets downloaded into the
        appropriate format for partitioning.
        """
        from simple_salesforce.exceptions import SalesforceMalformedRequest

        client = self.connector_config.get_client()

        ingest_docs = []
        for record_type in self.connector_config.categories:
            if record_type not in ACCEPTED_CATEGORIES:
                raise ValueError(f"{record_type} not currently an accepted Salesforce category")

            try:
                # Get ids from Salesforce
                records = client.query_all(
                    f"select Id from {record_type}",
                )
                for record in records["records"]:
                    ingest_docs.append(
                        SalesforceIngestDoc(
                            connector_config=self.connector_config,
                            processor_config=self.processor_config,
                            read_config=self.read_config,
                            record_type=record_type,
                            record_id=record["Id"],
                        ),
                    )
            except SalesforceMalformedRequest as e:
                raise SalesforceMalformedRequest(f"Problem with Salesforce query: {e}")

        return ingest_docs



from dataclasses import dataclass
from typing import TYPE_CHECKING

from pydantic import Field, Secret

from unstructured_ingest.v2.processes.connector_registry import (
    DestinationRegistryEntry,
    SourceRegistryEntry,
)
from unstructured_ingest.v2.processes.connectors.kafka.kafka import (
    KafkaAccessConfig,
    KafkaConnectionConfig,
    KafkaDownloader,
    KafkaDownloaderConfig,
    KafkaIndexer,
    KafkaIndexerConfig,
    KafkaUploader,
    KafkaUploaderConfig,
)

if TYPE_CHECKING:
    pass

CONNECTOR_TYPE = "kafka-local"


class LocalKafkaAccessConfig(KafkaAccessConfig):
    pass


class LocalKafkaConnectionConfig(KafkaConnectionConfig):
    access_config: Secret[LocalKafkaAccessConfig] = Field(
        default=LocalKafkaAccessConfig(), validate_default=True
    )

    def get_consumer_configuration(self) -> dict:
        bootstrap = self.bootstrap_server
        port = self.port

        conf = {
            "bootstrap.servers": f"{bootstrap}:{port}",
            "group.id": self.group_id,
            "enable.auto.commit": "false",
            "auto.offset.reset": "earliest",
        }
        return conf

    def get_producer_configuration(self) -> dict:
        bootstrap = self.bootstrap_server
        port = self.port

        conf = {
            "bootstrap.servers": f"{bootstrap}:{port}",
        }
        return conf


class LocalKafkaIndexerConfig(KafkaIndexerConfig):
    pass


@dataclass
class LocalKafkaIndexer(KafkaIndexer):
    connection_config: LocalKafkaConnectionConfig
    index_config: LocalKafkaIndexerConfig
    connector_type: str = CONNECTOR_TYPE


class LocalKafkaDownloaderConfig(KafkaDownloaderConfig):
    pass


@dataclass
class LocalKafkaDownloader(KafkaDownloader):
    connection_config: LocalKafkaConnectionConfig
    download_config: LocalKafkaDownloaderConfig
    connector_type: str = CONNECTOR_TYPE


class LocalKafkaUploaderConfig(KafkaUploaderConfig):
    pass


@dataclass
class LocalKafkaUploader(KafkaUploader):
    connection_config: LocalKafkaConnectionConfig
    upload_config: LocalKafkaUploaderConfig
    connector_type: str = CONNECTOR_TYPE


kafka_local_source_entry = SourceRegistryEntry(
    connection_config=LocalKafkaConnectionConfig,
    indexer=LocalKafkaIndexer,
    indexer_config=LocalKafkaIndexerConfig,
    downloader=LocalKafkaDownloader,
    downloader_config=LocalKafkaDownloaderConfig,
)

kafka_local_destination_entry = DestinationRegistryEntry(
    connection_config=LocalKafkaConnectionConfig,
    uploader=LocalKafkaUploader,
    uploader_config=LocalKafkaUploaderConfig,
)



import pytest

from unstructured_ingest.error import (
    DestinationConnectionError,
    PartitionError,
    SourceConnectionError,
)


@pytest.mark.parametrize(
    ("error_class", "exception_type", "error_message"),
    [
        (SourceConnectionError, ValueError, "Simulated connection error"),
        (DestinationConnectionError, RuntimeError, "Simulated connection error"),
        (PartitionError, FileNotFoundError, "Simulated partition error"),
    ],
)
def test_custom_error_decorator(error_class, exception_type, error_message):
    @error_class.wrap
    def simulate_error():
        raise exception_type(error_message)

    with pytest.raises(error_class) as context:
        simulate_error()

    expected_error_string = error_class.error_string.format(error_message)
    assert str(context.value) == expected_error_string



import json
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator

import pandas as pd
from pydantic import Field, Secret, model_validator

from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connector_registry import (
    DestinationRegistryEntry,
    SourceRegistryEntry,
)
from unstructured_ingest.v2.processes.connectors.sql.sql import (
    _DATE_COLUMNS,
    SQLAccessConfig,
    SqlBatchFileData,
    SQLConnectionConfig,
    SQLDownloader,
    SQLDownloaderConfig,
    SQLIndexer,
    SQLIndexerConfig,
    SQLUploader,
    SQLUploaderConfig,
    SQLUploadStager,
    SQLUploadStagerConfig,
    parse_date_string,
)

if TYPE_CHECKING:
    from sqlite3 import Connection as SqliteConnection
    from sqlite3 import Cursor as SqliteCursor

CONNECTOR_TYPE = "sqlite"


class SQLiteAccessConfig(SQLAccessConfig):
    pass


class SQLiteConnectionConfig(SQLConnectionConfig):
    access_config: Secret[SQLiteAccessConfig] = Field(
        default=SQLiteAccessConfig(), validate_default=True
    )
    database_path: Path = Field(
        description="Path to the .db file.",
    )

    @model_validator(mode="after")
    def check_database_path(self) -> "SQLiteConnectionConfig":
        if not self.database_path.exists():
            raise ValueError(f"{self.database_path} does not exist")
        if not self.database_path.is_file():
            raise ValueError(f"{self.database_path} is not a valid file")
        return self

    @contextmanager
    def get_connection(self) -> Generator["SqliteConnection", None, None]:
        from sqlite3 import connect

        connection = connect(database=self.database_path)
        try:
            yield connection
        finally:
            connection.commit()
            connection.close()

    @contextmanager
    def get_cursor(self) -> Generator["SqliteCursor", None, None]:
        with self.get_connection() as connection:
            cursor = connection.cursor()
            try:
                yield cursor
            finally:
                cursor.close()


class SQLiteIndexerConfig(SQLIndexerConfig):
    pass


@dataclass
class SQLiteIndexer(SQLIndexer):
    connection_config: SQLConnectionConfig
    index_config: SQLIndexerConfig
    connector_type: str = CONNECTOR_TYPE


class SQLiteDownloaderConfig(SQLDownloaderConfig):
    pass


@dataclass
class SQLiteDownloader(SQLDownloader):
    connection_config: SQLConnectionConfig
    download_config: SQLDownloaderConfig
    connector_type: str = CONNECTOR_TYPE
    values_delimiter: str = "?"

    def query_db(self, file_data: SqlBatchFileData) -> tuple[list[tuple], list[str]]:
        table_name = file_data.additional_metadata.table_name
        id_column = file_data.additional_metadata.id_column
        ids = [item.identifier for item in file_data.batch_items]
        with self.connection_config.get_connection() as sqlite_connection:
            cursor = sqlite_connection.cursor()
            fields = ",".join(self.download_config.fields) if self.download_config.fields else "*"
            values = ",".join(self.values_delimiter for _ in ids)
            query = f"SELECT {fields} FROM {table_name} WHERE {id_column} IN ({values})"
            logger.debug(f"running query: {query}\nwith values: {ids}")
            cursor.execute(query, ids)
            rows = cursor.fetchall()
            columns = [col[0] for col in cursor.description]
            return rows, columns


class SQLiteUploadStagerConfig(SQLUploadStagerConfig):
    pass


class SQLiteUploadStager(SQLUploadStager):
    upload_stager_config: SQLiteUploadStagerConfig


class SQLiteUploaderConfig(SQLUploaderConfig):
    pass


@dataclass
class SQLiteUploader(SQLUploader):
    upload_config: SQLiteUploaderConfig = field(default_factory=SQLiteUploaderConfig)
    connection_config: SQLiteConnectionConfig
    connector_type: str = CONNECTOR_TYPE

    def prepare_data(
        self, columns: list[str], data: tuple[tuple[Any, ...], ...]
    ) -> list[tuple[Any, ...]]:
        output = []
        for row in data:
            parsed = []
            for column_name, value in zip(columns, row):
                if isinstance(value, (list, dict)):
                    value = json.dumps(value)
                if column_name in _DATE_COLUMNS:
                    if value is None or pd.isna(value):
                        parsed.append(None)
                    else:
                        parsed.append(parse_date_string(value))
                else:
                    parsed.append(value)
            output.append(tuple(parsed))
        return output


sqlite_source_entry = SourceRegistryEntry(
    connection_config=SQLiteConnectionConfig,
    indexer_config=SQLiteIndexerConfig,
    indexer=SQLiteIndexer,
    downloader_config=SQLiteDownloaderConfig,
    downloader=SQLiteDownloader,
)

sqlite_destination_entry = DestinationRegistryEntry(
    connection_config=SQLiteConnectionConfig,
    uploader=SQLiteUploader,
    uploader_config=SQLiteUploaderConfig,
    upload_stager=SQLiteUploadStager,
    upload_stager_config=SQLiteUploadStagerConfig,
)



# https://developers.notion.com/reference/property-object#created-by
from dataclasses import dataclass, field
from typing import Optional

from htmlBuilder.tags import HtmlTag

from unstructured_ingest.connector.notion.interfaces import DBCellBase, DBPropertyBase
from unstructured_ingest.connector.notion.types.user import People


@dataclass
class CreatedBy(DBPropertyBase):
    id: str
    name: str
    type: str = "created_by"
    created_by: dict = field(default_factory=dict)

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class CreatedByCell(DBCellBase):
    id: str
    created_by: People
    type: str = "created_by"
    name: Optional[str] = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(created_by=People.from_dict(data.pop("created_by")), **data)

    def get_html(self) -> Optional[HtmlTag]:
        return self.created_by.get_html()



import base64
from pathlib import Path

from bs4 import BeautifulSoup
from pytest_mock import MockerFixture

from unstructured_ingest.utils.html import HtmlMixin
from unstructured_ingest.v2.interfaces import FileData, SourceIdentifiers


def test_extract_images(mocker: MockerFixture):
    mixin = HtmlMixin(extract_images=True)
    mock_download_response = b"DOWNLOADED"
    expected_image_src = base64.b64encode(mock_download_response).decode()
    mocked_download_response = mocker.patch(
        "unstructured_ingest.utils.html.HtmlMixin.download_content",
        return_value=mock_download_response,
    )
    url = "http://mywebsite.com/path/to/page"
    html = """
    <img src="http://mywebsite.com/img1.jpg"/>
    <img src="http://notmywebsite.com/img2.jpg"/>
    <img src="img3.jpg"/>
    <img src="data:image/png;base64,24689654..."/>
    """
    expected_html = f"""
    <img src="data:image/png;base64,{expected_image_src}"/>
    <img src="http://notmywebsite.com/img2.jpg"/>
    <img src="data:image/png;base64,{expected_image_src}"/>
    <img src="data:image/png;base64,24689654..."/>
    """
    expected_soup = BeautifulSoup(expected_html, "html.parser")
    result = mixin.extract_html_images(url=url, html=html)
    result_soup = BeautifulSoup(result, "html.parser")
    assert expected_soup == result_soup
    assert mocked_download_response.call_count == 2
    urls_to_download = [
        call_args_list.kwargs["url"] for call_args_list in mocked_download_response.call_args_list
    ]
    assert urls_to_download == ["http://mywebsite.com/img1.jpg", "http://mywebsite.com/img3.jpg"]


def test_extract_images_allow_list(mocker: MockerFixture):
    mixin = HtmlMixin(
        extract_images=True, allow_list=["http://allowedwebsite1.com", "http://allowedwebsite2.com"]
    )
    mock_download_response = b"DOWNLOADED"
    expected_image_src = base64.b64encode(mock_download_response).decode()
    mocked_download_response = mocker.patch(
        "unstructured_ingest.utils.html.HtmlMixin.download_content",
        return_value=mock_download_response,
    )
    url = "http://mywebsite.com/path/to/page"
    html = """
    <img src="http://mywebsite.com/img1.jpg"/>
    <img src="http://notmywebsite.com/img2.jpg"/>
    <img src="http://allowedwebsite1.com/img2.jpg"/>
    <img src="http://allowedwebsite2.com/img2.jpg"/>
    """

    expected_html = f"""
    <img src="http://mywebsite.com/img1.jpg"/>
    <img src="http://notmywebsite.com/img2.jpg"/>
    <img src="data:image/png;base64,{expected_image_src}"/>
    <img src="data:image/png;base64,{expected_image_src}"/>
    """
    expected_soup = BeautifulSoup(expected_html, "html.parser")
    result = mixin.extract_html_images(url=url, html=html)
    result_soup = BeautifulSoup(result, "html.parser")
    assert expected_soup == result_soup
    assert mocked_download_response.call_count == 2
    urls_to_download = [
        call_args_list.kwargs["url"] for call_args_list in mocked_download_response.call_args_list
    ]
    assert urls_to_download == [
        "http://allowedwebsite1.com/img2.jpg",
        "http://allowedwebsite2.com/img2.jpg",
    ]


def test_extract_embedded_docs(mocker: MockerFixture):
    mixin = HtmlMixin(extract_files=True)
    mock_download_response = b"DOWNLOADED"
    mocked_download_response = mocker.patch(
        "unstructured_ingest.utils.html.HtmlMixin.download_content",
        return_value=mock_download_response,
    )
    mocked_write_content = mocker.patch("unstructured_ingest.utils.html.HtmlMixin.write_content")
    url = "http://mywebsite.com/path/to/page"
    html = """
    <a href="http://mywebsite.com/file.pdf"/>
    <a href="http://notmywebsite.com/file.pdf"/>
    <a href="http://mywebsite.com/another/link"/>
    <a href="another/link/2"/>
    <a href="file.doc"/>
    """
    file_data = FileData(
        source_identifiers=SourceIdentifiers(
            fullpath="file.txt",
            filename="file.txt",
        ),
        connector_type="my_connector",
        identifier="mock_file_data",
    )
    results = mixin.extract_embedded_files(
        url=url, html=html, download_dir=Path("/tmp/download/location"), original_filedata=file_data
    )
    assert len(results) == 2
    downloaded_urls = [r["file_data"].metadata.url for r in results]
    assert downloaded_urls == ["http://mywebsite.com/file.pdf", "http://mywebsite.com/file.doc"]
    assert mocked_download_response.call_count == 2
    assert mocked_write_content.call_count == 2



import typing as t
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path

from unstructured_ingest.enhanced_dataclass import enhanced_field
from unstructured_ingest.error import SourceConnectionError, SourceConnectionNetworkError
from unstructured_ingest.interfaces import (
    AccessConfig,
    BaseConnectorConfig,
    BaseSingleIngestDoc,
    BaseSourceConnector,
    IngestDocCleanupMixin,
    SourceConnectorCleanupMixin,
    SourceMetadata,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.dep_check import requires_dependencies

if t.TYPE_CHECKING:
    from praw import Reddit


@dataclass
class RedditAccessConfig(AccessConfig):
    client_secret: t.Optional[str] = enhanced_field(default=None, sensitive=True)


@dataclass
class SimpleRedditConfig(BaseConnectorConfig):
    access_config: RedditAccessConfig
    subreddit_name: str
    num_posts: int
    user_agent: str
    client_id: str
    search_query: t.Optional[str] = None

    def __post_init__(self):
        if self.num_posts <= 0:
            raise ValueError("The number of Reddit posts to fetch must be positive.")


@dataclass
class RedditIngestDoc(IngestDocCleanupMixin, BaseSingleIngestDoc):
    connector_config: SimpleRedditConfig = field(repr=False)
    post_id: str
    registry_name: str = "reddit"

    def _create_full_tmp_dir_path(self):
        self.filename.parent.mkdir(parents=True, exist_ok=True)

    @SourceConnectionNetworkError.wrap
    @requires_dependencies(["praw"])
    def get_post(self):
        from praw import Reddit
        from praw.models import Submission

        reddit = Reddit(
            client_id=self.connector_config.client_id,
            client_secret=self.connector_config.access_config.client_secret,
            user_agent=self.connector_config.user_agent,
        )
        post = Submission(reddit, self.post_id)
        return post

    def update_source_metadata(self, **kwargs):
        post = kwargs.get("post", self.get_post())
        if post is None:
            self.source_metadata = SourceMetadata(
                exists=False,
            )
            return

        file_exists = (post.author != "[deleted]" or post.auth is not None) and (
            post.selftext != "[deleted]" or post.selftext != "[removed]"
        )

        self.source_metadata = SourceMetadata(
            date_created=datetime.utcfromtimestamp(post.created_utc).isoformat(),
            source_url=post.permalink,
            exists=file_exists,
        )

    @SourceConnectionError.wrap
    @BaseSingleIngestDoc.skip_if_file_exists
    def get_file(self):
        """Fetches the "remote" doc and stores it locally on the filesystem."""
        self._create_full_tmp_dir_path()
        # Write the title plus the body, if any
        post = self.get_post()
        self.update_source_metadata(post=post)
        if post is None:
            raise ValueError(
                f"Failed to retrieve post {self.post_id}",
            )

        text_to_write = f"# {post.title}\n{post.selftext}"
        with open(self.filename, "w", encoding="utf8") as f:
            f.write(text_to_write)

    @property
    def filename(self) -> Path:
        return (Path(self.read_config.download_dir) / f"{self.post_id}.md").resolve()

    @property
    def _output_filename(self):
        return Path(self.processor_config.output_dir) / f"{self.post_id}.json"

    @property
    def date_modified(self) -> t.Optional[str]:
        return None

    @property
    def version(self) -> t.Optional[str]:
        return None


@dataclass
class RedditSourceConnector(SourceConnectorCleanupMixin, BaseSourceConnector):
    connector_config: SimpleRedditConfig
    _reddit: t.Optional["Reddit"] = field(init=False, default=None)

    @property
    def reddit(self) -> "Reddit":
        from praw import Reddit

        if self._reddit is None:
            self._reddit = Reddit(
                client_id=self.connector_config.client_id,
                client_secret=self.connector_config.access_config.client_secret,
                user_agent=self.connector_config.user_agent,
            )
        return self._reddit

    @requires_dependencies(["praw"], extras="reddit")
    def initialize(self):
        _ = self.reddit

    def check_connection(self):
        from praw.endpoints import API_PATH
        from prawcore import ResponseException

        try:
            self.reddit._objectify_request(method="HEAD", params=None, path=API_PATH["me"])
        except ResponseException as response_error:
            logger.error(f"failed to validate connection: {response_error}", exc_info=True)
            raise SourceConnectionError(f"failed to validate connection: {response_error}")

    def get_ingest_docs(self):
        subreddit = self.reddit.subreddit(self.connector_config.subreddit_name)
        if self.connector_config.search_query:
            posts = subreddit.search(
                self.connector_config.search_query,
                limit=self.connector_config.num_posts,
            )
        else:
            posts = subreddit.hot(limit=self.connector_config.num_posts)
        return [
            RedditIngestDoc(
                connector_config=self.connector_config,
                processor_config=self.processor_config,
                read_config=self.read_config,
                post_id=post.id,
            )
            for post in posts
        ]



import typing as t
from dataclasses import dataclass

from unstructured_ingest.enhanced_dataclass import EnhancedDataClassJsonMixin
from unstructured_ingest.interfaces import BaseDestinationConnector
from unstructured_ingest.runner.writers.base_writer import Writer

if t.TYPE_CHECKING:
    from unstructured_ingest.connector.chroma import ChromaWriteConfig, SimpleChromaConfig


@dataclass
class ChromaWriter(Writer, EnhancedDataClassJsonMixin):
    write_config: "ChromaWriteConfig"
    connector_config: "SimpleChromaConfig"

    def get_connector_cls(self) -> t.Type[BaseDestinationConnector]:
        from unstructured_ingest.connector.chroma import (
            ChromaDestinationConnector,
        )

        return ChromaDestinationConnector
